"""Mixin classes for basis."""

from __future__ import annotations

import abc
import copy
import inspect
import re
from collections import OrderedDict
from contextlib import contextmanager
from functools import wraps
from itertools import chain
from typing import TYPE_CHECKING, Any, Generator, Literal, Optional, Tuple, Union

import jax
import numpy as np
from numpy.typing import ArrayLike, NDArray
from pynapple import Tsd, TsdFrame, TsdTensor

from ..convolve import create_convolutional_predictor
from ..utils import _get_terminal_size, format_repr
from ._composition_utils import (
    _composite_basis_setter_logic,
    _recompute_all_default_labels,
    generate_basis_label_pair,
    generate_composite_basis_labels,
    get_input_shape,
    infer_input_dimensionality,
    is_basis_like,
    label_setter,
    set_input_shape,
)
from ._transformer_basis import TransformerBasis

if TYPE_CHECKING:
    from ._basis import Basis


def _is_basis(object: Any):
    """Minimal checks for attributes."""
    return all(
        hasattr(object, attrname) for attrname in ("compute_features", "get_params")
    )


def set_input_shape_state(states: Tuple[str] = ("_input_shape_product",)):
    """
    Preserve input shape-related attributes during method execution.

    This decorator ensures that the attributes `_n_basis_input_` and `_input_shape_product`
    are copied from the original object (`self`) to the returned object (`klass`)
    after the wrapped method executes. It is intended to be used with methods that
    clone or create a new instance of the class, ensuring these critical attributes
    are retained for functionality such as cross-validation.

    Parameters
    ----------
    method :
        The method to be wrapped. This method is expected to return an object
        (`klass`) that requires the `_n_basis_input_` and `_input_shape_` attributes.
    attr_list

    Returns
    -------
    :
        The wrapped method that copies `_input_shape_product` and `_input_shape_` from
        the original object (`self`) to the new object (`klass`).

    Examples
    --------
    Applying the decorator to a method:

    >>> from functools import wraps
    >>> @set_input_shape_state
    ... def __sklearn_clone__(self):
    ...     klass = self.__class__(**self.get_params())
    ...     return klass

    The `_input_shape_product` and `_input_shape_` attributes of `self` will be
    copied to `klass` after the method executes.
    """

    def decorator(method):
        @wraps(method)
        def wrapper(self, *args, **kwargs):
            # Call the original method
            klass = method(self, *args, **kwargs)
            # Copy the specified attributes
            for attr_name in states:
                setattr(klass, attr_name, getattr(self, attr_name, None))
            return klass

        return wrapper

    return decorator


def remap_parameters(method):
    """Remap parameter names to original."""

    @wraps(method)
    def wrapper(self, **params):
        # get key/param map
        _, map_params = self._map_parameters()

        # use mapped key if exists, or original key
        # deepcopy basis to avoid matching labels when assigning twice the same basis
        # to two different components:
        # basis.set_params(basis1=bas, basis2=bas)
        new_params = {
            map_params.get(key, key): (copy.deepcopy(val) if _is_basis(val) else val)
            for key, val in params.items()
        }

        # apply set_params
        self = method(self, **new_params)

        # re-assign ordered ids to all default labels of atomic components
        # note that the recursion always run until parent.
        if self._parent is None:
            _recompute_all_default_labels(self)

        return self

    return wrapper


class BasisMixin:

    def __init__(self, label: Optional[str] = None):
        if not hasattr(self, "_input_shape_"):
            self._input_shape_ = None
        # initialize as default
        self._label = self.__class__.__name__
        # pass through the setter
        self.label = label
        # initialize parent to None. This should not end in "_" because it is
        # a permanent property of a basis, defined at composite basis init
        self._parent: Optional["BasisMixin"] = None

    def __repr__(self):
        return format_repr(self)

    def __getitem__(self, index: str) -> Basis:

        if isinstance(index, (int, slice)):
            string = "Slicing" if isinstance(index, slice) else "Indexing with integer"
            raise IndexError(
                f"You can only index basis using labels. {string} is invalid."
            )

        search = next(
            (bas for lab, bas in generate_basis_label_pair(self) if lab == index), None
        )

        if search is None:
            avail_index = ",".join(f"'{b}'" for b in self._generate_subtree_labels())
            raise IndexError(
                f"Basis label {index} not found. Available labels: {avail_index}"
            )
        return search

    def _generate_subtree_labels(
        self, type_label: Literal["all", "user-defined"] = "all"
    ) -> Generator[str]:
        """List all user-specified labels."""
        yield from generate_composite_basis_labels(self, type_label)

    def _iterate_over_components(self) -> Generator | chain:
        """Return a generator that iterates over all basis components.

        For atomic bases, the list is just [self].

        Returns
        -------
            A generator returning self, it will be chained in composite bases.

        """
        if hasattr(self, "basis1") and hasattr(self, "basis2"):
            components1 = (
                self.basis1._iterate_over_components()
                if hasattr(self.basis1, "_iterate_over_components")
                else [self.basis1]
            )
            components2 = (
                self.basis2._iterate_over_components()
                if hasattr(self.basis2, "_iterate_over_components")
                else [self.basis2]
            )
            return chain(
                components1,
                components2,
            )
        return (x for x in [self])

    @property
    def label(self) -> str:
        """Label for the basis."""
        if getattr(self, "_label", None) is None and hasattr(self, "_generate_label"):
            return self._generate_label()
        return getattr(self, "_label", self.__class__.__name__)

    @label.setter
    def label(self, label: str | None) -> None:
        error = label_setter(self, label)
        if error:
            raise error

    def set_input_shape(
        self,
        *xi: int | tuple[int, ...] | NDArray,
        allow_inputs_of_different_shape: bool = True,
    ) -> BasisMixin:
        """Set the expected input shape for the basis object."""
        if getattr(self, "_parent", None) is not None:
            raise ValueError(
                "Cannot set input shape on a child basis. Set the input shape on the root basis instead.\n"
                "For example, instead of ``self.basis1.set_input_shape(n); self.basis2.set_input_shape(m)``, "
                "do ``self.set_input_shape(n, m)``."
            )
        set_input_shape(
            self, *xi, allow_inputs_of_different_shape=allow_inputs_of_different_shape
        )
        return self

    @property
    def input_shape(self):
        """
        Expected per-sample input shape.

        Returns
        -------
        :
            If inputs are shaped ``(n_samples, *shape)``, returns ``shape``.
        """
        return get_input_shape(self)

    def _get_feature_slicing(
        self,
        n_inputs: Optional[tuple] = None,
        start_slice: Optional[int] = None,
    ) -> Tuple[OrderedDict, int]:
        """
        Calculate and return the slicing for features based on the input structure.

        This method determines how to slice the features for different basis types.

        Parameters
        ----------
        n_inputs :
            The number of input basis for each component, by default it uses ``self._input_shape_product``.
        start_slice :
            The starting index for slicing, by default it starts from 0.

        Returns
        -------
        split_dict :
            Dictionary with keys as labels and values as slices representing
            the slicing for each additive component.
        start_slice :
            The updated starting index after slicing.

        See Also
        --------
        _get_default_slicing : Handles default slicing logic.
        _merge_slicing_dicts : Merges multiple slicing dictionaries, handling keys conflicts.
        """
        # Set default values for start_slice if not provided
        start_slice = start_slice or 0
        # Handle the default case for non-additive basis types
        # See overwritten method for recursion logic
        split_dict, start_slice = self._get_default_slicing(start_slice=start_slice)

        return split_dict, start_slice

    def _get_default_slicing(self, start_slice: int) -> Tuple[OrderedDict, int]:
        """Handle default slicing logic."""
        split_dict = {
            self.label: slice(start_slice, start_slice + self.n_output_features)
        }
        start_slice += self.n_output_features
        return OrderedDict(split_dict), start_slice

    def split_by_feature(
        self,
        x: NDArray,
        axis: int = 1,
    ):
        r"""
        Decompose an array along a specified axis into sub-arrays based on the number of expected inputs.

        This function takes an array (e.g., a design matrix or model coefficients) and splits it along
        a designated axis.

        **How it works:**

        - If the basis expects an input shape ``(n_samples, n_inputs)``, then the feature axis length will
          be ``total_n_features = n_inputs * n_basis_funcs``. This axis is reshaped into dimensions
          ``(n_inputs, n_basis_funcs)``.

        - If the basis expects an input of shape ``(n_samples,)``, then the feature axis length will
          be ``total_n_features = n_basis_funcs``. This axis is reshaped into ``(1, n_basis_funcs)``.

        For example, if the input array ``x`` has shape ``(1, 2, total_n_features, 4, 5)``,
        then after applying this method, it will be reshaped into ``(1, 2, n_inputs, n_basis_funcs, 4, 5)``.

        The specified axis (``axis``) determines where the split occurs, and all other dimensions
        remain unchanged. See the example section below for the most common use cases.

        Parameters
        ----------
        x :
            The input array to be split, representing concatenated features, coefficients,
            or other data. The shape of ``x`` along the specified axis must match the total
            number of features generated by the basis, i.e., ``self.n_output_features``.

            **Examples:**

            - For a design matrix: ``(n_samples, total_n_features)``

            - For model coefficients: ``(total_n_features,)`` or ``(total_n_features, n_neurons)``.

        axis : int, optional
            The axis along which to split the features. Defaults to 1.
            Use ``axis=1`` for design matrices (features along columns) and ``axis=0`` for
            coefficient arrays (features along rows). All other dimensions are preserved.

        Raises
        ------
        ValueError
            If the shape of ``x`` along the specified axis does not match ``self.n_output_features``.

        Returns
        -------
        dict
            A dictionary where:

            - **Key**: Label of the basis.
            - **Value**: the array reshaped to: ``(..., n_inputs, n_basis_funcs, ...)``
        """
        # convert axis to positive ints
        axis = axis if axis >= 0 else x.ndim + axis

        if x.shape[axis] != self.n_output_features:
            raise ValueError(
                "`x.shape[axis]` does not match the expected number of features."
                f" `x.shape[axis] == {x.shape[axis]}`, while the expected number "
                f"of features is {self.n_output_features}"
            )

        # Get the slice dictionary based on predefined feature slicing
        slice_dict = self._get_feature_slicing()[0]

        # Helper function to build index tuples for each slice
        def build_index_tuple(slice_obj, axis: int, ndim: int):
            """Create an index tuple to apply a slice on the given axis."""
            index = [slice(None)] * ndim  # Initialize index for all dimensions
            index[axis] = slice_obj  # Replace the axis with the slice object
            return tuple(index)

        # Get the dict for slicing the correct axis
        index_dict = jax.tree_util.tree_map(
            lambda sl: build_index_tuple(sl, axis, x.ndim), slice_dict
        )

        # Custom leaf function to identify index tuples as leaves
        def is_leaf(value):
            # Check if it's a tuple, length matches ndim, and all elements are slice objects
            if isinstance(value, tuple) and len(value) == x.ndim:
                return all(isinstance(v, slice) for v in value)
            return False

        # Apply the slicing using the custom leaf function
        out = jax.tree_util.tree_map(lambda sl: x[sl], index_dict, is_leaf=is_leaf)
        reshaped_out: dict = dict()
        for items, bas in zip(out.items(), self):
            key, val = items
            reshaped_out[key] = bas._reshape_concatenated_arrays(val, bas, axis)
        return reshaped_out

    @staticmethod
    def _reshape_concatenated_arrays(
        array: NDArray, bas: BasisMixin, axis: int
    ) -> NDArray:
        # reshape the arrays to match input shapes
        shape = list(array.shape)
        array = array.reshape(
            shape[:axis]
            + [*(b for sh in bas._input_shape_ for b in sh), -1]
            + shape[axis + 1 :]
        )
        return array

    def __iter__(self):
        """Make basis iterable. Re-implemented for additive."""
        yield self


class AtomicBasisMixin(BasisMixin):
    """Mixin class for atomic bases (i.e. non-composite)."""

    _is_complex = False

    def __init__(self, n_basis_funcs: int, label: Optional[str] = None):
        super().__init__(label=label)
        self._n_basis_funcs = n_basis_funcs
        check_basis_min = getattr(self, "_check_n_basis_min", None)
        if check_basis_min:
            check_basis_min()

    @property
    def is_complex(self):
        """
        Whether the basis is intrinsically complex.

        Returns
        -------
        :
            ``True`` if the basis is complex; ``False`` otherwise.

        Notes
        -----
        :meth:`compute_features` always returns a real-valued design matrix. For
        complex bases (e.g., ``FourierEval``), the real and imaginary parts are
        returned as separate columns.
        """
        return self.__class__._is_complex

    @property
    def n_output_features(self) -> int | None:
        """
        Number of features returned by the basis.

        Notes
        -----
        The number of output features can be determined only when the number of inputs
        provided to the basis is known. Therefore, before the first call to ``compute_features``,
        this property will return ``None``. After that call, or after setting the input shape with
        ``set_input_shape``, ``n_output_features`` will be available.
        """
        if self._input_shape_product is not None:
            return self.n_basis_funcs * self._input_shape_product[0]
        return None

    @property
    def _has_default_label(self):
        return re.match(rf"^{self.__class__.__name__}(_\d+)?$", self._label) is not None

    @set_input_shape_state(states=("_input_shape_product", "_input_shape_", "_label"))
    def __sklearn_clone__(self) -> Basis:
        """Clone the basis while preserving attributes related to input shapes.

        This method ensures that input shape attributes (e.g., `_input_shape_product`,
        `_input_shape_`) are preserved during cloning. Reinitializing the class
        as in the regular sklearn clone would drop these attributes, rendering
        cross-validation unusable.
        """
        klass = self.__class__(**self.get_params())
        return klass

    def set_input_shape(self, xi: int | tuple[int, ...] | NDArray) -> BasisMixin:
        """
        Set the expected input shape for the basis object.

        This method configures the shape of the input data that the basis object expects.
        ``xi`` can be specified as an integer, a tuple of integers, or derived
        from an array. The method also calculates the total number of input
        features and output features based on the number of basis functions.

        Parameters
        ----------
        xi :
            The input shape specification.
            - An integer: Represents the dimensionality of the input. A value of ``1`` is treated as scalar input.
            - A tuple: Represents the exact input shape excluding the first axis (sample axis).
              All elements must be integers.
            - An array: The shape is extracted, excluding the first axis (assumed to be the sample axis).

        Raises
        ------
        ValueError
            If a tuple is provided and it contains non-integer elements.

        Returns
        -------
        self :
            Returns the instance itself to allow method chaining.

        Notes
        -----
        All state attributes that depends on the input must be set in this method in order for
        the API of basis to work correctly. In particular, this method is called by ``setup_basis``,
        which is equivalent to ``fit`` for a transformer. If any input dependent state
        is not set in this method, then ``compute_features`` (equivalent to ``fit_transform``) will break.

        """
        return super().set_input_shape(xi)


class EvalBasisMixin:
    """Mixin class for evaluational basis."""

    def __init__(self, bounds: Optional[Tuple[float, float]] = None):
        self.bounds = bounds

    def _compute_features(self, *xi: ArrayLike | Tsd | TsdFrame | TsdTensor):
        """Evaluate basis at sample points.

        The basis is evaluated at the locations specified in the inputs. For example,
        ``compute_features(np.array([0, .5]))`` would return the array:

        .. code-block:: text

           b_1(0) ... b_n(0)
           b_1(.5) ... b_n(.5)

        where ``b_i`` is the i-th basis.

        Parameters
        ----------
        *xi:
            The input samples over which to apply the basis transformation. The samples can be passed
            as multiple arguments, each representing a different dimension for multivariate inputs.

        Returns
        -------
        :
            A matrix with the transformed features.

        """
        out = self.evaluate(*(np.reshape(x, (x.shape[0], -1)) for x in xi))
        return np.reshape(out, (out.shape[0], -1))

    def setup_basis(self, *xi: NDArray) -> Basis:
        """
        Set all basis states.

        This method corresponds sklearn transformer ``fit``. As fit, it must receive the input and
        it must set all basis states, i.e. ``kernel_`` and all the states relative to the input shape.
        The difference between this method and the transformer ``fit`` is in the expected input structure,
        where the transformer ``fit`` method requires the inputs to be concatenated in a 2D array, while here
        each input is provided as a separate time series for each basis element.

        Parameters
        ----------
        xi:
            Input arrays.

        Returns
        -------
        :
            The basis with ready for evaluation.
        """
        self.set_input_shape(*xi)
        return self

    def _set_input_independent_states(self) -> "EvalBasisMixin":
        """
        Compute all the basis states that do not depend on the input.

        For EvalBasisMixin, this method might not perform any operation but simply return the
        instance itself, as no kernel preparation is necessary.

        Returns
        -------
        self :
            The instance itself.

        """
        return self

    @property
    def bounds(self):
        """Range of values covered by the basis."""
        return self._bounds

    @staticmethod
    def _format_bounds(values: Any) -> Tuple[Any, Exception | None]:
        """Check bounds and cast to tuple."""

        if not hasattr(values, "__len__"):
            raise TypeError(
                "Invalid bounds provided. ``bounds`` must be a tuple of floats."
                f"``bounds`` {values} of type {type(values)} provided instead."
            )

        elif values is not None and len(values) != 2:
            raise ValueError(
                f"The provided `bounds` must be of length two. The bounds ``{values}`` have length "
                f"{len(values)} instead!"
            )

        try:
            values = values if values is None else tuple(map(float, values))
        except (ValueError, TypeError) as e:
            raise TypeError(
                "Could not convert `bounds` to float. "
                f"The provided bounds values are '{values}'."
            ) from e

        if values is not None and values[1] <= values[0]:
            raise ValueError(
                f"Invalid bound {values}. Lower bound is greater or equal than the upper bound."
            )

        return values

    @bounds.setter
    def bounds(self, values: Union[None, Tuple[float, float]]):
        """Setter for bounds."""
        if values is None:
            self._bounds = None
            return
        values = self._format_bounds(values)
        if values is not None and len(values) != 2:
            raise ValueError(
                f"The provided `bounds` must be of length two. Length {len(values)} provided instead!"
            )
        self._bounds = values


class ConvBasisMixin:
    """Mixin class for convolutional basis."""

    def __init__(self, window_size: int, conv_kwargs: Optional[dict] = None):
        self.kernel_ = None
        self.window_size = window_size
        self.conv_kwargs = {} if conv_kwargs is None else conv_kwargs

    def _compute_features(self, *xi: NDArray | Tsd | TsdFrame | TsdTensor):
        """Convolve basis functions with input time series.

        A bank of basis filters is convolved with the input data. All the dimensions
        except for the sample-axis are flattened, so that the method always returns a
        matrix.

        For example, if inputs are of shape (num_samples, 2, 3), the output will be
        ``(num_samples, num_basis_funcs * 2 * 3)``.

        Parameters
        ----------
        *xi:
            The input data over which to apply the basis transformation. The samples can be passed
            as multiple arguments, each representing a different dimension for multivariate inputs.

        Notes
        -----
        This method is intended to be 1-to-1 mappable to sklearn ``transform`` method of transformer. This
        means that for the method to be callable, all the state attributes have to be pre-computed in a
        method that is mappable to ``fit``, which for us is ``_fit_basis``. It is fundamental that both
        methods behaves like the corresponding transformer method, with the only difference being the input
        structure: a single (X, y) pair for the transformer, a number of time series for the Basis.

        """
        self._check_has_kernel()
        # before calling the convolve, check that the input matches
        # the expectation. We can check xi[0] only, since convolution
        # is applied at the end of the recursion on the 1D basis, ensuring len(xi) == 1.
        conv = create_convolutional_predictor(self.kernel_, *xi, **self._conv_kwargs)
        # make sure to return a matrix
        return np.reshape(conv, (conv.shape[0], -1))

    def setup_basis(self, *xi: NDArray) -> Basis:
        """
        Set all basis states.

        This method corresponds sklearn transformer ``fit``. As fit, it must receive the input and
        it must set all basis states, i.e. ``kernel_`` and all the states relative to the input shape.
        The difference between this method and the transformer ``fit`` is in the expected input structure,
        where the transformer ``fit`` method requires the inputs to be concatenated in a 2D array, while here
        each input is provided as a separate time series for each basis element.

        Parameters
        ----------
        xi:
            Input arrays.

        Returns
        -------
        :
            The basis with ready for evaluation.
        """
        self._set_kernel()
        self.set_input_shape(*xi)
        return self

    def _set_input_independent_states(self):
        """
        Compute all the basis states that do not depend on the input.

        For Conv mixin the only attribute is the kernel.
        """
        return self._set_kernel()

    def _set_kernel(self) -> "ConvBasisMixin":
        """
        Prepare or compute the convolutional kernel for the basis functions.

        This method is called to prepare the basis functions for convolution operations
        in subclasses. It computes a
        kernel based on the basis functions that will be used for convolution with the
        input data. The specifics of kernel computation depend on the subclass implementation
        and the nature of the basis functions.

        Returns
        -------
        self :
            The instance itself, modified to include the computed kernel. This
            allows for method chaining and integration into transformation pipelines.

        Notes
        -----
        Subclasses implementing this method should detail the specifics of how the kernel is
        computed and how the input parameters are utilized.

        """
        self.kernel_ = self.evaluate(np.linspace(0, 1, self.window_size))
        return self

    @property
    def window_size(self):
        """Duration of the convolutional kernel in number of samples."""
        return self._window_size

    @window_size.setter
    def window_size(self, window_size):
        """Setter overwrite default setter the window size parameter."""
        self._check_window_size(window_size)
        self._window_size = window_size

    def _check_window_size(self, window_size):
        if window_size is None:
            raise ValueError("You must provide a window_size!")

        elif not (isinstance(window_size, int) and window_size > 0):
            raise ValueError(
                f"`window_size` must be a positive integer. {window_size} provided instead!"
            )

    @property
    def conv_kwargs(self):
        """The convolutional kwargs.

        Keyword arguments passed to :func:`nemos.convolve.create_convolutional_predictor`.
        """
        return self._conv_kwargs

    @conv_kwargs.setter
    def conv_kwargs(self, values: dict):
        """Check and set convolution kwargs."""
        self._check_convolution_kwargs(values)
        self._conv_kwargs = values

    @staticmethod
    def _check_convolution_kwargs(conv_kwargs: dict):
        """Check convolution kwargs settings.

        Raises
        ------
        ValueError:
            If ``axis`` is provided as an argument, and it is different from 0
            (samples must always be in the first axis).
        ValueError:
            If ``self._conv_kwargs`` include parameters not recognized or that do not have
            default values in ``create_convolutional_predictor``.
        """
        if "axis" in conv_kwargs:
            raise ValueError(
                "Setting the `axis` parameter is not allowed. Basis requires the "
                "convolution to be applied along the first axis (`axis=0`).\n"
                "Please transpose your input so that the desired axis for "
                "convolution is the first dimension (axis=0)."
            )
        convolve_params = inspect.signature(create_convolutional_predictor).parameters
        convolve_configs = {
            key
            for key, param in convolve_params.items()
            if param.default
            # prevent user from passing
            # `basis_matrix` or `time_series` in kwargs.
            is not inspect.Parameter.empty
        }
        if not set(conv_kwargs.keys()).issubset(convolve_configs):
            # do not encourage to set axis.
            convolve_configs = convolve_configs.difference({"axis"})
            # remove the parameter in case axis=0 was passed, since it is allowed.
            invalid = (
                set(conv_kwargs.keys())
                .difference(convolve_configs)
                .difference({"axis"})
            )
            raise ValueError(
                f"Unrecognized keyword arguments: {invalid}. "
                f"Allowed convolution keyword arguments are: {convolve_configs}."
            )

    def _check_has_kernel(self) -> None:
        """Check that the kernel is pre-computed."""
        if self.kernel_ is None:
            raise RuntimeError(
                "You must call `setup_basis` before `_compute_features` for Conv basis."
            )


class BasisTransformerMixin:
    """Mixin class for constructing a transformer."""

    def to_transformer(self) -> TransformerBasis:
        """
        Turn the Basis into a TransformerBasis for use with scikit-learn.

        Examples
        --------
        Jointly cross-validating basis and GLM parameters with scikit-learn.

        >>> import nemos as nmo
        >>> from sklearn.pipeline import Pipeline
        >>> from sklearn.model_selection import GridSearchCV
        >>> # load some data
        >>> X, y = np.random.normal(size=(30, 1)), np.random.poisson(size=30)
        >>> basis = nmo.basis.RaisedCosineLinearEval(10).set_input_shape(1).to_transformer()
        >>> glm = nmo.glm.GLM(regularizer="Ridge", regularizer_strength=1.)
        >>> pipeline = Pipeline([("basis", basis), ("glm", glm)])
        >>> param_grid = dict(
        ...     glm__regularizer_strength=(0.1, 0.01, 0.001, 1e-6),
        ...     basis__n_basis_funcs=(3, 5, 10, 20, 100),
        ... )
        >>> gridsearch = GridSearchCV(
        ...     pipeline,
        ...     param_grid=param_grid,
        ...     cv=5,
        ... )
        >>> gridsearch = gridsearch.fit(X, y)
        """
        return TransformerBasis(self)


class CompositeBasisMixin(BasisMixin):
    """Mixin class for composite basis.

    Add overwrites concrete methods or defines abstract methods for composite basis
    (AdditiveBasis and MultiplicativeBasis).
    """

    _shallow_copy: bool = False

    def __init__(
        self, basis1: BasisMixin, basis2: BasisMixin, label: Optional[str] = None
    ):
        # number of input arrays that the basis receives
        self._n_input_dimensionality = infer_input_dimensionality(
            basis1
        ) + infer_input_dimensionality(basis2)

        # set the attribute
        self._is_complex = getattr(basis1, "is_complex", False) or getattr(
            basis2, "is_complex", False
        )

        # This step is slow if you add a very large number of bases
        self._basis1 = None
        self._basis2 = None
        # deep copy to avoid changes directly to the 1d basis to be reflected
        # in the composite basis.

        if not self.__class__._shallow_copy:
            basis1 = copy.deepcopy(basis1)
            basis2 = copy.deepcopy(basis2)

        self.basis1 = basis1
        self.basis2 = basis2

        # set parents
        self.basis1._parent = self
        self.basis2._parent = self

        # trigger label setter
        super().__init__(label=label)

    @property
    def is_complex(self):
        # is_complex is used to check if basis can be multiplied with another
        # allowed multiplications: real x real, real x complex.
        return self._is_complex

    def _is_basis_like(
        self, basis1: Optional[BasisMixin] = None, basis2: Optional[BasisMixin] = None
    ):
        if basis1 and not is_basis_like(basis1):
            raise ValueError(
                "`basis1` does not implement `compute_features`. "
                "The method is required for the correct behavior of the basis."
            )
        if basis2 and not is_basis_like(basis2):
            raise ValueError(
                "`basis2` does not implement `compute_features`. "
                "The method is required for the correct behavior of the basis."
            )

    @property
    def basis1(self):
        """Return first component."""
        return self._basis1

    @basis1.setter
    def basis1(self, basis):
        self._is_basis_like(basis1=basis)

        if self._basis2:
            self._set_labels(basis, self._basis2)
        if self._basis1:
            basis = _composite_basis_setter_logic(basis, self._basis1)
        self._basis1 = basis
        self._input_shape_update()

    @property
    def basis2(self):
        """Return second component."""
        return self._basis2

    @basis2.setter
    def basis2(self, basis):
        self._is_basis_like(basis2=basis)
        if self._basis1:
            self._set_labels(self._basis1, basis)
        if self._basis2:
            basis = _composite_basis_setter_logic(basis, self._basis2)
        self._basis2 = basis
        self._input_shape_update()

    @property
    def _has_default_label(self):
        return self._label is None

    def _check_unique_labels(self, basis1, basis2):
        """Check that all user-defined labels in the given basis objects are unique."""
        # Include self's label in uniqueness check (if applicable)
        self_label = getattr(self, "_label", None)

        # Store basis1 labels
        seen_labels = set(basis1._generate_subtree_labels("user-defined"))
        if self_label in seen_labels:
            err_msg = (
                f"All user-provided labels of basis elements must be distinct.\n"
                f"The basis you are composing share the following labels: '{self_label}'.\n"
                "Please change the labels for one of the elements before composition."
            )
            return True, err_msg

        # Check for duplicates in basis2 without storing in set for efficiency
        for label in basis2._generate_subtree_labels("user-defined"):
            if label == self_label or label in seen_labels:
                err_msg = (
                    f"All user-provided labels of basis elements must be distinct.\n"
                    f"The basis you are composing share the following labels: '{label}'.\n"
                    "Please change the labels for one of the elements before composition."
                )
                return True, err_msg

        return False, ""

    def _set_labels(self, basis1, basis2):
        # check labels
        non_unique, err_msg = self._check_unique_labels(basis1, basis2)
        if non_unique:
            raise ValueError(err_msg)

        self.update_default_label_id(basis1, basis2)

    def _input_shape_update(self):
        # if all bases where set, then set input for composition.
        set_bases = (s is not None for s in self.input_shape)

        if all(set_bases):
            # pass down the input shapes
            self.set_input_shape(*self.input_shape)

    @property
    def _input_shape_(self):
        return self.input_shape

    def set_input_shape(
        self, *xi: int | tuple[int, ...] | NDArray, allow_inputs_of_different_shape=True
    ) -> BasisMixin:
        """
        Set the expected input shape for the basis object.

        This method sets the input shape for each component basis in the basis.
        One ``xi`` must be provided for each basis component, specified as an integer,
        a tuple of integers, or an array. The method calculates and stores the total number of output features
        based on the number of basis functions in each component and the provided input shapes.

        Parameters
        ----------
        *xi :
            The input shape specifications. For every k,``xi[k]`` can be:
            - An integer: Represents the dimensionality of the input. A value of ``1`` is treated as scalar input.
            - A tuple: Represents the exact input shape excluding the first axis (sample axis).
              All elements must be integers.
            - An array: The shape is extracted, excluding the first axis (assumed to be the sample axis).

        allow_inputs_of_different_shape :
            True if the composition allows input of different shape (as in addition), False otherwise
            (as in multiplication).

        Raises
        ------
        ValueError
            If a tuple is provided, and it contains non-integer elements.
            If not enough inputs are provided.

        Returns
        -------
        self :
            Returns the instance itself to allow method chaining.
        """
        return super().set_input_shape(
            *xi, allow_inputs_of_different_shape=allow_inputs_of_different_shape
        )

    @property
    @abc.abstractmethod
    def n_basis_funcs(self):
        """Read only property for composite bases."""
        pass

    def setup_basis(self, *xi: NDArray) -> Basis:
        """
        Set all basis states.

        This method corresponds sklearn transformer ``fit``. As fit, it must receive the input and
        it must set all basis states, i.e. ``kernel_`` and all the states relative to the input shape.
        The difference between this method and the transformer ``fit`` is in the expected input structure,
        where the transformer ``fit`` method requires the inputs to be concatenated in a 2D array, while here
        each input is provided as a separate time series for each basis element.

        Parameters
        ----------
        xi:
            Input arrays.

        Returns
        -------
        :
            The basis with ready for evaluation.
        """
        # setup both input independent
        self._set_input_independent_states()

        # and input dependent states
        self.set_input_shape(*xi)

        return self

    def _set_input_independent_states(self):
        """
        Compute the input dependent states for traversing the composite basis.

        Returns
        -------
        :
            The basis with the states stored as attributes of each component.
        """
        if hasattr(self.basis1, "_set_input_independent_states"):
            self.basis1._set_input_independent_states()
        if hasattr(self.basis2, "_set_input_independent_states"):
            self.basis2._set_input_independent_states()

    @contextmanager
    def _set_shallow_copy(self, value):
        """Context manager for setting the shallow copy flag in a thread safe way."""
        old_value = self.__class__._shallow_copy
        self.__class__._shallow_copy = value
        try:
            yield
        finally:
            self.__class__._shallow_copy = old_value

    @set_input_shape_state(states=("_input_shape_product", "_label"))
    def __sklearn_clone__(self) -> Basis:
        """Clone the basis while preserving attributes related to input shapes.

        This method ensures that input shape attributes (e.g., `_input_shape_product`,
        `_input_shape_`) are preserved during cloning. Reinitializing the class
        as in the regular sklearn clone would drop these attributes, rendering
        cross-validation unusable.
        The method also handles recursive cloning for composite basis structures.

        Notes
        -----
        The ``_shallow_copy`` attribute is set to True in the context, forcing a shallow copy, at
        before the klass definition, and reset to False after cloning.
        """
        with self._set_shallow_copy(True):
            # clone recursively
            basis1 = self.basis1.__sklearn_clone__()
            basis2 = self.basis2.__sklearn_clone__()

            # shallow copy init
            klass = self.__class__(basis1, basis2)

        return klass

    def __repr__(self, n=0):
        """Nested repr for composite basis."""
        cols, rows = _get_terminal_size()
        rows = rows // 4
        cols = cols
        disp_label = len(str(self.label)) < cols
        if disp_label:
            start_str = f"'{self.label}': "
        else:
            start_str = ""

        # number of nested composite bases
        n += 1
        tab = "    "
        try:
            basis1 = self.basis1.__repr__(n=n)
        except TypeError:
            basis1 = self.basis1
        try:
            basis2 = self.basis2.__repr__(n=n)
        except TypeError:
            basis2 = self.basis2
        if n < rows:
            rep = (
                start_str + f"{self.__class__.__name__}"
                f"(\n{n*tab}basis1={basis1},\n{n*tab}basis2={basis2},\n{(n-1)*tab})"
            )
        elif n == rows:
            rep = start_str + f"{self.__class__.__name__}(\n{n*tab}...\n{(n-1)*tab})"
        else:
            rep = None
        return rep

    @staticmethod
    def update_default_label_id(basis1: Basis, basis2: Basis):
        """
        Update basis2 atomic element labels.

        When composing bases, update basis2 labels in order to disambiguate them with respect to basis1 labels.
        Here we assume that each tree (basis1 and basis2) have unique basis labels if taken separately, while
        they may have overlapping labels between them.
        This function updates the label of basis2 to avoid overlaps with basis1.


        Parameters
        ----------
        basis1:
            The first basis object.
        basis2:
            The second basis object.

        Notes
        -----
        The method only operates on default labels (i.e. class names).
        User-defined labels will not be edited, and overlaps will result in `ValueError` at composite basis
        initialization.

        """
        delta_labels = dict()
        for bas in basis1._iterate_over_components():
            cls_name = bas.__class__.__name__
            if cls_name not in delta_labels:
                pattern = re.compile(rf"^{cls_name}(_\d+)?$")
                delta_labels[cls_name] = sum(
                    (
                        1
                        for b in basis1._iterate_over_components()
                        if re.match(pattern, b._label)
                    )
                )

        # update the label
        count_labels = dict()
        for bas in basis2._iterate_over_components():
            cls_name = bas.__class__.__name__
            if cls_name not in delta_labels:
                continue
            match = re.match(rf"^{cls_name}(_\d+)?$", bas._label)
            if match:
                current = count_labels.get(cls_name, 0)
                bas._label = (
                    f"{cls_name}_{current + delta_labels[cls_name]}"
                    if current + delta_labels[cls_name]
                    else cls_name
                )
                count_labels[cls_name] = 1 + current

        return

    def __sklearn_get_params__(self, deep=True):
        """
        Implement standard scikit-learn get parameters by inspecting init.

        This function will be called by Base.set_params() to get the actual param
        structure, since the inherited``get_params`` is overridden to use basis labels
        instead of the nested structure.

        Parameters
        ----------
        deep:
            If true, call itself recursively on basis implementing the method.

        Returns
        -------
        out:
            A dictionary containing the parameters. Key is the parameter
            name, value is the parameter value.
        """
        out = dict()
        for key in self._get_param_names():
            value = getattr(self, key)
            if (
                deep
                and hasattr(value, "__sklearn_get_params__")
                and not isinstance(value, type)
            ):
                item_orig = value.__sklearn_get_params__()
                deep_items = item_orig.items()
                out.update((key + "__" + k, val) for k, val in deep_items)
            out[key] = value
        return out

    def _map_parameters(self, deep=True):
        """
        Remap parameters in a given object by replacing 'basis[12]' patterns with unique labels.

        Parameters
        ----------
        bas : object
            An object that contains a `get_params()` method returning a dictionary of parameters.

        Returns
        -------
        param_dict_map:
            A dictionary with renamed parameters.
        key_map:
            A mapping of new parameter names to original names.

        """
        param_dict_map, key_map = self._get_params_and_key_map(
            deep=deep,
        )  # Retrieve the parameter dictionary
        # strip higher level label
        param_dict_map = self._remove_self_label_from_key(param_dict_map)
        key_map = self._remove_self_label_from_key(key_map)
        return param_dict_map, key_map

    def _get_params_and_key_map(self, deep=True) -> Tuple[dict, dict]:
        """
        From scikit-learn, get parameters by inspecting init.

        Parameters
        ----------
        deep

        Returns
        -------
        parameter_dict:
            A dictionary containing the parameters. Key is the ``basis_label "__" + parameter_name``,
            value is the parameter value.
        key_map:
            Dictionary that maps the keys of parameter_dict onto the keys based on attribute nesting.
        """
        parameter_dict = dict()
        key_map = dict()
        for key in self._get_param_names():
            value = getattr(self, key)
            if deep and not isinstance(value, type):
                if hasattr(value, "_get_params_and_key_map"):
                    item_map, key_mapping = value._get_params_and_key_map()
                    map_deep_items = item_map.items()
                    # only keep the last basis label (the leaf of the tree)
                    parameter_dict.update(
                        (
                            (
                                "__".join(k.split("__")[-2:])
                                if not _is_basis(val)
                                else val.label
                            ),
                            val,
                        )
                        for k, val in map_deep_items
                    )
                    key_map.update(
                        {
                            k: key + "__" + v
                            for k, v in zip(item_map.keys(), key_mapping.values())
                        }
                    )
                elif hasattr(value, "get_params"):
                    # hit this on atomic bases
                    deep_items = value.get_params().items()
                    parameter_dict.update(
                        {f"{value.label}__{k}": val for k, val in deep_items}
                    )
                    # assume that the basis is either 1 or 2
                    lab = "basis1" if value is self.basis1 else "basis2"
                    key_map.update(
                        {f"{value.label}__{k}": lab + "__" + k for k, _ in deep_items}
                    )

            if _is_basis(value):
                parameter_dict[value.label] = value
                key_map[value.label] = key
            else:
                parameter_dict[self.label + "__" + key] = value
                key_map[self.label + "__" + key] = key
        return parameter_dict, key_map

    def _remove_self_label_from_key(self, mapping_dict: dict) -> dict:
        initial_string = self.label + "__"
        return {
            k[len(initial_string) :] if k.startswith(initial_string) else k: val
            for k, val in mapping_dict.items()
        }

    def get_params(self, deep=True) -> dict:
        """Get parameters using labels."""
        new_param_dict, _ = self._map_parameters(deep=deep)
        return new_param_dict

    @remap_parameters
    def set_params(self, **params: Any):
        """Map parameters using labels and set."""
        return super().set_params(**params)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from absl.testing import absltest
from absl.testing import parameterized

import jax
import jax.numpy as jnp

from nemos.third_party.jaxopt.jaxopt import projection
from nemos.third_party.jaxopt.jaxopt import prox
from nemos.third_party.jaxopt.jaxopt import tree_util
from nemos.third_party.jaxopt.jaxopt._src import test_util

import numpy as onp


class ProxTest(test_util.JaxoptTestCase):

  def test_prox_none(self):
    rng = onp.random.RandomState(0)
    x = rng.rand(20) * 2 - 1
    self.assertArraysAllClose(prox.prox_none(x), x)

  # A scalar implementation for check purposes.
  def _prox_l1(self, x, alpha):
    if x >= alpha:
      return x - alpha
    elif x <= -alpha:
      return x + alpha
    else:
      return 0

  def test_prox_lasso(self):
    rng = onp.random.RandomState(0)

    # Check forward pass with array x and scalar alpha.
    x = rng.rand(20) * 2 - 1
    alpha = 0.5
    expected = jnp.array([self._prox_l1(x[i], alpha) for i in range(len(x))])
    got = prox.prox_lasso(x, alpha)
    self.assertArraysAllClose(expected, got)

    # Check computed Jacobian against manual Jacobian.
    jac = jax.jacobian(prox.prox_lasso)(x, alpha)
    jac_exact = onp.zeros_like(jac)
    for i in range(len(x)):
      if x[i] >= alpha:
        jac_exact[i, i] = 1
      elif x[i] <= -alpha:
        jac_exact[i, i] = 1
      else:
        jac_exact[i, i] = 0
    self.assertArraysAllClose(jac_exact, jac)

    # Check forward pass with array x and array alpha.
    alpha = rng.rand(20)
    expected = jnp.array([self._prox_l1(x[i], alpha[i]) for i in range(len(x))])
    got = prox.prox_lasso(x, alpha)
    self.assertArraysAllClose(expected, got)

    # Check forward pass with pytree x and pytree alpha.
    x = (rng.rand(20) * 2 - 1, rng.rand(20) * 2 - 1)
    alpha = (rng.rand(20), rng.rand(20))
    expected0 = [self._prox_l1(x[0][i], alpha[0][i]) for i in range(len(x[0]))]
    expected1 = [self._prox_l1(x[1][i], alpha[1][i]) for i in range(len(x[0]))]
    expected = (jnp.array(expected0), jnp.array(expected1))
    got = prox.prox_lasso(x, alpha)
    self.assertArraysAllClose(jnp.array(expected), jnp.array(got))

    # Check forward pass with pytree x and tuple-of-scalars alpha.
    alpha = (0.5, 0.2)
    expected0 = [self._prox_l1(x[0][i], alpha[0]) for i in range(len(x[0]))]
    expected1 = [self._prox_l1(x[1][i], alpha[1]) for i in range(len(x[0]))]
    expected = (jnp.array(expected0), jnp.array(expected1))
    got = prox.prox_lasso(x, alpha)
    self.assertArraysAllClose(jnp.array(expected), jnp.array(got))

    # test that works when regularizer is float and prox jit compiled
    got = jax.jit(prox.prox_lasso)(x, 0.5)
    expected0 = [self._prox_l1(x[0][i], 0.5) for i in range(len(x[0]))]
    expected1 = [self._prox_l1(x[1][i], 0.5) for i in range(len(x[0]))]
    expected = (jnp.array(expected0), jnp.array(expected1))
    self.assertArraysAllClose(jnp.array(expected), jnp.array(got))

  def _prox_enet(self, x, lam, gamma):
    return (1.0 / (1.0 + lam * gamma)) * self._prox_l1(x, lam)

  def test_prox_elastic_net(self):
    rng = onp.random.RandomState(0)

    # Check forward pass with array x and scalar alpha.
    x = rng.rand(20) * 2 - 1
    hyperparams = (0.5, 0.1)
    expected = jnp.array([self._prox_enet(x[i], *hyperparams)
                          for i in range(len(x))])
    got = prox.prox_elastic_net(x, hyperparams)
    self.assertArraysAllClose(expected, got)

    # Check forward pass with array x and array alpha.
    hyperparams = (rng.rand(20), rng.rand(20))
    expected = jnp.array([self._prox_enet(x[i], hyperparams[0][i],
                                          hyperparams[1][i])
                          for i in range(len(x))])
    got = prox.prox_elastic_net(x, hyperparams)
    self.assertArraysAllClose(expected, got)

    # Check forward pass with pytree x.
    x = (rng.rand(20) * 2 - 1, rng.rand(20) * 2 - 1)
    hyperparams = (0.5, 0.1)
    expected0 = [self._prox_enet(x[0][i], *hyperparams)
                 for i in range(len(x[0]))]
    expected1 = [self._prox_enet(x[1][i], *hyperparams)
                 for i in range(len(x[0]))]
    expected = (jnp.array(expected0), jnp.array(expected1))
    got = prox.prox_elastic_net(x, ((0.5, 0.5), (0.1, 0.1)))
    self.assertArraysAllClose(jnp.array(expected), jnp.array(got))

  # A pure NumPy implementation for check purposes.
  def _prox_l2(self, x, alpha):
    l2_norm = onp.sqrt(onp.sum(x ** 2))
    return max(1 - alpha / l2_norm, 0) * x

  def test_prox_group_lasso(self):
    rng = onp.random.RandomState(0)
    x = rng.rand(20) * 2 - 1

    # An example with non-zero block.
    alpha = 0.1
    got = prox.prox_group_lasso(x, alpha)
    expected = self._prox_l2(x, alpha)

    # An example with zero block.
    alpha = 10.0
    self.assertArraysAllClose(got, expected)
    got = prox.prox_group_lasso(x, alpha)
    expected = self._prox_l2(x, alpha)

  def test_prox_ridge(self):
    rng = onp.random.RandomState(0)
    x = rng.rand(20) * 2 - 1
    x = jnp.array(x)
    alpha = 10.0

    # The objective solved by the prox is differentiable.
    # We can check that the gradient is zero.
    def fun(y):
      """f(y) = 0.5 ||y - x||^2 + 0.5 * alpha ||y||_2^2."""
      diff = x - y
      return 0.5 * jnp.sum(diff ** 2) + 0.5 * alpha * jnp.sum(y ** 2)

    got = prox.prox_ridge(x, alpha)
    self.assertArraysAllClose(jax.grad(fun)(got), jnp.zeros_like(got))

  def test_prox_non_negative_ridge(self):
    rng = onp.random.RandomState(0)
    x = rng.rand(20) * 2 - 1
    x = jnp.array(x)
    alpha = 10.0

    def fun(y):
      """f(y) = 0.5 ||y - x||^2 + 0.5 * alpha ||y||_2^2."""
      diff = x - y
      return 0.5 * jnp.sum(diff ** 2) + 0.5 * alpha * jnp.sum(y ** 2)

    got = prox.prox_non_negative_ridge(x, alpha)
    fixed_point = jax.nn.relu(got - jax.grad(fun)(got))
    self.assertArraysAllClose(got, fixed_point)

  def test_prox_non_negative_lasso(self):
    rng = onp.random.RandomState(0)
    x = rng.rand(20) * 2 - 1
    x = jnp.array(x)
    alpha = 0.5

    def fun(y):
      """f(y) = 0.5 ||y - x||^2 + alpha * sum(y)."""
      diff = x - y
      return 0.5 * jnp.sum(diff ** 2) + alpha * jnp.sum(y)

    got = prox.prox_non_negative_lasso(x, alpha)
    fixed_point = jax.nn.relu(got - jax.grad(fun)(got))
    self.assertArraysAllClose(got, fixed_point)

  def test_make_prox_from_projection(self):
    rng = onp.random.RandomState(0)
    x = rng.rand(10)
    proxop = prox.make_prox_from_projection(projection.projection_simplex)
    self.assertArraysAllClose(proxop(x), projection.projection_simplex(x))

  @parameterized.product(
      prox_op=[
          prox.prox_none,
          prox.prox_lasso,
          prox.prox_non_negative_lasso,
          prox.prox_ridge,
          prox.prox_non_negative_ridge,
          prox.prox_elastic_net,
      ]
  )
  def test_pytree_comptability(self, prox_op):
    rng = onp.random.RandomState(0)
    x = dict(a=rng.rand(16, 16), b=rng.rand(16))
    got = prox_op(x)
    expected = tree_util.tree_map(prox_op, x)
    self.assertAllClose(got, expected)
    if prox_op is prox.prox_lasso:
      l1_reg = tree_util.tree_ones_like(x)
      got = prox_op(x, l1_reg)
      expected = tree_util.tree_map(prox_op, x, l1_reg)
      self.assertAllClose(got, expected)
    if prox_op is prox.prox_elastic_net:
      hyperparams = [tree_util.tree_ones_like(x), tree_util.tree_ones_like(x)]
      got = prox_op(x, hyperparams)
      hyperparams_tree = tree_util.tree_map(
          lambda y: [jnp.ones_like(y), jnp.ones_like(y)], x)
      expected = tree_util.tree_map(prox_op, x, hyperparams_tree)
      self.assertAllClose(got, expected)

if __name__ == '__main__':
  absltest.main()



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Linear system solvers."""

from typing import Any
from typing import Callable
from typing import Optional
from typing import Tuple

import jax
import jax.numpy as jnp
import jax.scipy as jsp

from nemos.third_party.jaxopt.jaxopt._src.tree_util import tree_add_scalar_mul


def _materialize_array(matvec, shape, dtype=None):
  """Materializes the matrix A used in matvec(x) = Ax."""
  x = jnp.zeros(shape, dtype)
  return jax.jacfwd(matvec)(x)


def _make_ridge_matvec(matvec: Callable, ridge: float = 0.0):
  def ridge_matvec(v: Any) -> Any:
    return tree_add_scalar_mul(matvec(v), ridge, v)
  return ridge_matvec


def solve_lu(matvec: Callable, 
             b: jnp.ndarray, 
             ridge: Optional[float] = None) -> jnp.ndarray:
  """Solves ``A x = b`` using ``jax.lax.solve``.

  This solver is based on an LU decomposition.
  It will materialize the matrix ``A`` in memory.

  Args:
    matvec: product between ``A`` and a vector.
    b: array.
    ridge: optional ridge regularization.

  Returns:
    array with same structure as ``b``.
  """
  if ridge is not None:
    matvec = _make_ridge_matvec(matvec, ridge=ridge)
  A = _materialize_array(matvec, b.shape, b.dtype) # 4d array (tensor) if len(b.shape) == 2
  if len(b.shape) == 0:
    if A == 0.0:
      raise ValueError('A should be non-zero for scalar b.')
    return b / A
  elif len(b.shape) == 1:
    return jax.numpy.linalg.solve(A, b)
  elif len(b.shape) == 2:
    A = A.reshape(-1, b.shape[0] * b.shape[1])  # 2d array (matrix)
    return jax.numpy.linalg.solve(A, b.ravel()).reshape(*b.shape)
  else:
    raise NotImplementedError


def solve_cholesky(matvec: Callable,
                   b: jnp.ndarray,
                   ridge: Optional[float] = None) -> jnp.ndarray:
  """Solves ``A x = b``, using Cholesky decomposition.

  It will materialize the matrix ``A`` in memory.

  Args:
    matvec: product between positive definite matrix ``A`` and a vector.
    b: array.
    ridge: optional ridge regularization.

  Returns:
    array with same structure as ``b``.
  """
  if ridge is not None:
    matvec = _make_ridge_matvec(matvec, ridge=ridge)
  A = _materialize_array(matvec, b.shape)
  if len(b.shape) == 0:
    if A == 0.0:
      raise ValueError('A should be non-zero for scalar b.')
    return b / A
  elif len(b.shape) == 1:
    return jsp.linalg.solve(A, b, assume_a='pos')
  elif len(b.shape) == 2:
    return  jsp.linalg.solve(A, b.ravel(), assume_a='pos').reshape(*b.shape)
  else:
    raise NotImplementedError


def solve_inv(matvec: Callable,
              b: jnp.ndarray,
              ridge: Optional[float] = None) -> jnp.ndarray:
  """Solves ``A x = b``, using matrix inversion.

  It will materialize the matrix ``A`` in memory.

  Args:
    matvec: product between positive definite matrix ``A`` and a vector.
    b: array.
    ridge: optional ridge regularization.

  Returns:
    array with same structure as ``b``.
  """
  if ridge is not None:
    matvec = _make_ridge_matvec(matvec, ridge=ridge)
  A = _materialize_array(matvec, b.shape)
  if len(b.shape) == 0:
    if A == 0.0:
      raise ValueError('A should be non-zero for scalar b.')
    return b / A
  elif len(b.shape) == 1:
    A = _materialize_array(matvec, b.shape)
    return jnp.dot(jnp.linalg.inv(A), b)
  else:
    raise NotImplementedError


def solve_qr(matvec: Callable, 
             b: jnp.ndarray, 
             ridge: Optional[float] = None) -> Tuple[jnp.ndarray, jnp.ndarray]:
  """Solves ``A x = b``, using QR factorization.

  It will materialize the matrix ``A`` in memory.

  Args:
    matvec: product between matrix ``A`` and a vector.
    b: RHS array.
    ridge: optional ridge regularization.

  Returns:
    A tuple containing Q and R matrices.
  """
  if ridge is not None:
    matvec = _make_ridge_matvec(matvec, ridge=ridge)
  A = _materialize_array(matvec, b.shape)
  if len(b.shape) == 0:
    if A == 0.0:
      raise ValueError('A should be non-zero for scalar b.')
    return b / A
  elif len(b.shape) == 1:
    # TODO: consider passing mode='economic' instead of current 'full'.
    q, r = jsp.linalg.qr(A)
    return jsp.linalg.solve_triangular(r, q.T @ b)
  else:
    # TODO: support for 2D b arrays.
    raise NotImplementedError

def solve_cg(matvec: Callable,
             b: Any,
             ridge: Optional[float] = None,
             init: Optional[Any] = None,
             **kwargs) -> Any:
  """Solves ``A x = b`` using conjugate gradient.

  It assumes that ``A`` is  a Hermitian, positive definite matrix.

  Args:
    matvec: product between ``A`` and a vector.
    b: pytree.
    ridge: optional ridge regularization.
    init: optional initialization to be used by conjugate gradient.
    **kwargs: additional keyword arguments for solver.

  Returns:
    pytree with same structure as ``b``.
  """
  if ridge is not None:
    matvec = _make_ridge_matvec(matvec, ridge=ridge)
  return jsp.sparse.linalg.cg(matvec, b, x0=init, **kwargs)[0]


def _make_rmatvec(matvec, x):
  transpose = jax.linear_transpose(matvec, x)
  return lambda y: transpose(y)[0]


def _normal_matvec(matvec, x):
  """Computes A^T A x from matvec(x) = A x."""
  matvec_x, vjp = jax.vjp(matvec, x)
  return vjp(matvec_x)[0]


def solve_normal_cg(matvec: Callable,
                    b: Any,
                    ridge: Optional[float] = None,
                    init: Optional[Any] = None,
                    **kwargs) -> Any:
  """Solves the normal equation ``A^T A x = A^T b`` using conjugate gradient.

  This can be used to solve Ax=b using conjugate gradient when A is not
  hermitian, positive definite.

  Args:
    matvec: product between ``A`` and a vector.
    b: pytree.
    ridge: optional ridge regularization.
    init: optional initialization to be used by normal conjugate gradient.
    **kwargs: additional keyword arguments for solver.

  Returns:
    pytree with same structure as ``b``.
  """
  if init is None:
    example_x = b  # This assumes that matvec is a square linear operator.
  else:
    example_x = init

  try:
    rmatvec = _make_rmatvec(matvec, example_x)
  except TypeError:
    raise TypeError("The initialization `init` of solve_normal_cg is "
                    "compulsory when `matvec` is nonsquare. It should "
                    "have the same pytree structure as a solution. "
                    "Typically, a pytree filled with zeros should work.")

  def normal_matvec(x):
    return _normal_matvec(matvec, x)

  if ridge is not None:
    normal_matvec = _make_ridge_matvec(normal_matvec, ridge=ridge)

  Ab = rmatvec(b)  # A.T b

  return jsp.sparse.linalg.cg(normal_matvec, Ab, x0=init, **kwargs)[0]


def solve_gmres(matvec: Callable,
                b: Any,
                ridge: Optional[float] = None,
                init: Optional[Any] = None,
                tol: float = 1e-5,
                **kwargs) -> Any:
  """Solves ``A x = b`` using gmres.

  Args:
    matvec: product between ``A`` and a vector.
    b: pytree.
    ridge: optional ridge regularization.
    init: optional initialization to be used by gmres.
    **kwargs: additional keyword arguments for solver.

  Returns:
    pytree with same structure as ``b``.
  """
  if ridge is not None:
    matvec = _make_ridge_matvec(matvec, ridge=ridge)
  return jsp.sparse.linalg.gmres(matvec, b, tol=tol, x0=init, **kwargs)[0]


def solve_bicgstab(matvec: Callable,
                   b: Any,
                   ridge: Optional[float] = None,
                   init: Optional[Any] = None,
                   **kwargs) -> Any:
  """Solves ``A x = b`` using bicgstab.

  Args:
    matvec: product between ``A`` and a vector.
    b: pytree.
    ridge: optional ridge regularization.
    init: optional initialization to be used by bicgstab.
    **kwargs: additional keyword arguments for solver.

  Returns:
    pytree with same structure as ``b``.
  """
  if ridge is not None:
    matvec = _make_ridge_matvec(matvec, ridge=ridge)
  return jsp.sparse.linalg.bicgstab(matvec, b, x0=init, **kwargs)[0]



"""Solvers wrapping Optax solvers with Optimistix for use with NeMoS."""

import abc
import inspect
from typing import Any, Callable, ClassVar, NamedTuple, Union

import equinox as eqx
import jax
import jax.numpy as jnp
import optax
import optimistix as optx

from ..regularizer import Regularizer
from ..tree_utils import tree_sub
from ..typing import Pytree
from ._optimistix_solvers import (
    DEFAULT_ATOL,
    DEFAULT_MAX_STEPS,
    DEFAULT_RTOL,
    OptimistixAdapter,
    OptimistixStepResult,
    Params,
)


class AbstractOptimistixOptaxSolver(OptimistixAdapter, abc.ABC):
    """Adapter for optimistix.OptaxMinimiser which is an adapter for Optax solvers."""

    _solver_cls = optx.OptaxMinimiser
    # if defined, the docstring is extended to include the documentation of the wrapped Optax solver
    _optax_solver: ClassVar[Callable[..., optax.GradientTransformation]]

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)

        # only append things if the _optax_solver class attribute is defined
        if not hasattr(cls, "_optax_solver"):
            return

        doc_so_far = inspect.cleandoc(inspect.getdoc(cls))
        # delete the part about OptaxMinimiser
        doc_so_far = doc_so_far.split("\n\nOptaxMinimiser's documentation:", 1)[0]

        init_header = inspect.cleandoc(f"More info from {cls.__name__}.__init__'s doc")
        init_header += "\n" + "-" * len(init_header)
        init_doc = inspect.cleandoc(
            inspect.getdoc(cls.__init__)
            or f"No documentation found for {cls.__name__}.init"
        )
        init_doc = init_header + "\n" + init_doc

        optax_header = inspect.cleandoc(
            f"""
            More info from Optax's {cls._optax_solver.__name__} documentation:
            """
        )
        optax_header += "\n" + "-" * len(optax_header)
        optax_doc = inspect.cleandoc(
            inspect.getdoc(cls._optax_solver) or "No documentation found in Optax."
        )
        optax_doc = optax_header + "\n" + optax_doc

        full_doc = "\n\n".join(
            (
                doc_so_far,
                init_doc,
                optax_doc,
            )
        )

        cls.__doc__ = inspect.cleandoc(full_doc)


class ScaleByLearningRateState(NamedTuple):
    learning_rate: Union[float, jax.Array]


def stateful_scale_by_learning_rate(
    stepsize: float, flip_sign: bool = True
) -> optax.GradientTransformation:
    """
    Reimplementation of optax.scale_by_learning_rate, just storing the learning rate in the state.

    Required for setting the scaling appropriately when used with
    proximal gradient descent.
    """
    m = -1 if flip_sign else 1

    def init_fn(params):
        del params
        return ScaleByLearningRateState(jnp.array(stepsize))

    def update_fn(updates, state, params=None):
        del params
        updates = jax.tree.map(lambda g: m * stepsize * g, updates)

        return updates, state

    return optax.GradientTransformation(init_fn, update_fn)  # pyright: ignore


def _make_rate_scaler(
    stepsize: float | None,
    linesearch_kwargs: dict[str, Any] | None,
) -> optax.GradientTransformation:
    """
    Make an Optax transformation for setting the learning rate.

    If `stepsize` is not None and larger than 0, use it as a constant learning rate.
    Otherwise `optax.scale_by_zoom_linesearch` is used with `linesearch_kwargs`.
    By default the curvature condition is disabled, which reduces to a backtracking
    linesearch where the stepsizes are chosen using the cubic or quadratic interpolation
    used in zoom linesearch.
    By default, 15 linesearch steps are used, which can be overwritten with
    `max_linesearch_steps` in `linesearch_kwargs`.
    """
    if stepsize is None or stepsize <= 0.0:
        if linesearch_kwargs is None:
            linesearch_kwargs = {}

        if "max_linesearch_steps" not in linesearch_kwargs:
            linesearch_kwargs["max_linesearch_steps"] = 15

        if "curv_rtol" not in linesearch_kwargs:
            linesearch_kwargs["curv_rtol"] = jnp.inf

        return optax.scale_by_zoom_linesearch(**linesearch_kwargs)
    else:
        if linesearch_kwargs:
            raise ValueError("Only provide stepsize or linesearch_kwargs.")
        # GradientDescent works with optax.scale_by_learning_rate as well
        # but for ProximalGradient we need to be able to extract the current learning rate
        return stateful_scale_by_learning_rate(stepsize)


class OptimistixOptaxGradientDescent(AbstractOptimistixOptaxSolver):
    """
    Gradient descent implementation combining Optax and Optimistix.

    Uses Optax's SGD combined with Optax's zoom linesearch or a constant learning rate.

    The full optimization loop is handled by the `optimistix.OptaxMinimiser` wrapper.
    """

    _optax_solver = optax.sgd

    def __init__(
        self,
        unregularized_loss: Callable,
        regularizer: Regularizer,
        regularizer_strength: float | None,
        tol: float = DEFAULT_ATOL,
        rtol: float = DEFAULT_RTOL,
        maxiter: int = DEFAULT_MAX_STEPS,
        momentum: float | None = None,
        acceleration: bool = True,
        stepsize: float | None = None,
        linesearch_kwargs: dict | None = None,
        **solver_init_kwargs,
    ):
        """
        Create a solver wrapping `optax.sgd`.

        If `acceleration` is True, use the Nesterov acceleration as defined by Sutskever et al. 2013.
        Note that this is different from the Nesterov acceleration implemented by JAXopt and
        only has an effect if `momentum` is used as well.

        If `stepsize` is not None and larger than 0, use it as a constant learning rate.
        Otherwise `optax.scale_by_zoom_linesearch` is used with the curvature condition
        disabled, which reduces to a backtracking linesearch where the stepsizes are chosen
        using the cubic or quadratic interpolation used in zoom linesearch.
        By default, 15 linesearch steps are used, which can be overwritten with
        `max_linesearch_steps` in `linesearch_kwargs`.

        References
        ----------
        [1] [Sutskever, I., Martens, J., Dahl, G. &amp; Hinton, G.. (2013).
        "On the importance of initialization and momentum in deep learning."
        Proceedings of the 30th International Conference on Machine Learning, PMLR 28(3):1139-1147, 2013.
        ](https://proceedings.mlr.press/v28/sutskever13.html)
        """
        _sgd = optax.chain(
            optax.sgd(learning_rate=1.0, momentum=momentum, nesterov=acceleration),
            _make_rate_scaler(stepsize, linesearch_kwargs),
        )
        solver_init_kwargs["optim"] = _sgd

        super().__init__(
            unregularized_loss,
            regularizer,
            regularizer_strength,
            tol=tol,
            rtol=rtol,
            maxiter=maxiter,
            **solver_init_kwargs,
        )

    @classmethod
    def get_accepted_arguments(cls) -> set[str]:
        arguments = super().get_accepted_arguments()

        arguments.discard("optim")  # we create this, it can't be passed

        return arguments

    @classmethod
    def _note_about_accepted_arguments(cls) -> str:
        note = super()._note_about_accepted_arguments()
        accel_nesterov = inspect.cleandoc(
            """
            `acceleration` is passed to `optax.sgd` as the `nesterov` parameter.
            Note that this only has an effect if `momentum` is used as well.
            """
        )
        return inspect.cleandoc(note + "\n" + accel_nesterov)


class OptimistixOptaxProximalGradient(AbstractOptimistixOptaxSolver):
    """
    ProximalGradient implementation combining Optax and Optimistix.

    Uses Optax's SGD combined with Optax's zoom linesearch or a constant learning rate.
    Then uses the learning rate given by Optax to scale the proximal
    operator's update and check for convergence using Optimistix's criterion.

    Works with the same proximal operator functions as JAXopt did.
    """

    fun: Callable
    fun_with_aux: Callable
    prox: Callable

    # stats: dict[str, PyTree[ArrayLike]]
    stats: dict[str, Pytree]

    _proximal: ClassVar[bool] = True

    def __init__(
        self,
        unregularized_loss: Callable,
        regularizer: Regularizer,
        regularizer_strength: float | None,
        tol: float = DEFAULT_ATOL,
        rtol: float = DEFAULT_RTOL,
        maxiter: float = DEFAULT_MAX_STEPS,
        momentum: float | None = None,
        acceleration: bool = True,
        stepsize: float | None = None,
        linesearch_kwargs: dict | None = None,
        **solver_init_kwargs,
    ):
        """
        Create a proximal gradient solver using `optax.sgd` and applying the proximal operator on each update step.

        If `acceleration` is True, use the Nesterov acceleration as defined by Sutskever et al. 2013.
        Note that this is different from the Nesterov acceleration implemented by JAXopt and
        only has an effect if `momentum` is used as well.

        If `stepsize` is not None and larger than 0, use it as a constant learning rate.
        Otherwise `optax.scale_by_zoom_linesearch` is used with the curvature condition
        disabled, which reduces to a backtracking linesearch where the stepsizes are chosen
        using the cubic or quadratic interpolation used in zoom linesearch.
        By default, 15 linesearch steps are used, which can be overwritten with
        `max_linesearch_steps` in `linesearch_kwargs`.

        References
        ----------
        [1] [Sutskever, I., Martens, J., Dahl, G. &amp; Hinton, G.. (2013).
        "On the importance of initialization and momentum in deep learning."
        Proceedings of the 30th International Conference on Machine Learning, PMLR 28(3):1139-1147, 2013.
        ](https://proceedings.mlr.press/v28/sutskever13.html)
        """
        _sgd = optax.chain(
            optax.sgd(learning_rate=1.0, momentum=momentum, nesterov=acceleration),
            _make_rate_scaler(stepsize, linesearch_kwargs),
        )
        solver_init_kwargs["optim"] = _sgd

        super().__init__(
            unregularized_loss,
            regularizer,
            regularizer_strength,
            tol=tol,
            rtol=rtol,
            maxiter=maxiter,
            **solver_init_kwargs,
        )

    @classmethod
    def get_accepted_arguments(cls) -> set[str]:
        arguments = super().get_accepted_arguments()

        arguments.discard("optim")  # we create this, it can't be passed

        return arguments

    def get_learning_rate(self, state: optx._solver.optax._OptaxState) -> float:
        """
        Read out the learning rate for scaling within the proximal operator.

        This learning rate is either a static learning rate or was found by a linesearch.
        """
        return state.opt_state[-1].learning_rate

    def step(
        self,
        fn: Callable,
        y: Params,
        args: Pytree,
        options: dict[str, Any],
        state: optx._solver.optax._OptaxState,
        tags: frozenset[object],
    ):
        # take gradient step
        new_params, new_state, new_aux = self._solver.step(
            fn, y, args, options, state, tags
        )

        # apply the proximal operator
        new_params = self.prox(
            new_params,
            self.regularizer_strength,
            self.get_learning_rate(new_state),
        )

        # reevaluate function value at the new point
        new_state = eqx.tree_at(lambda s: s.f, new_state, fn(new_params, args)[0])

        # recheck convergence criteria with the projected point
        updates = tree_sub(new_params, y)

        # replicating the jaxopt stopping criterion
        terminate = (
            optx.two_norm(updates) / self.get_learning_rate(new_state)
            < self._solver.atol
        )

        new_state = eqx.tree_at(lambda s: s.terminate, new_state, terminate)

        return new_params, new_state, new_aux

    def run(
        self,
        init_params: Params,
        *args,
    ) -> OptimistixStepResult:
        solution = optx.minimise(
            fn=self.fun,
            solver=self,  # pyright: ignore
            y0=init_params,
            args=args,
            options=self.config.options,
            has_aux=self.config.has_aux,
            max_steps=self.config.maxiter,
            adjoint=self.config.adjoint,
            throw=self.config.throw,
            tags=self.config.tags,
        )

        self.stats.update(solution.stats)

        return solution.value, solution.state

    def init(self, *args, **kwargs):
        # so that when passing self to optx.minimise, init can be called
        return self._solver.init(*args, **kwargs)

    def terminate(self, *args, **kwargs):
        # so that when passing self to optx.minimise, terminate can be called
        return self._solver.terminate(*args, **kwargs)

    def postprocess(self, *args, **kwargs):
        # so that when passing self to optx.minimise, postprocess can be called
        return self._solver.postprocess(*args, **kwargs)

    @classmethod
    def _note_about_accepted_arguments(cls) -> str:
        note = super()._note_about_accepted_arguments()
        accel_nesterov = inspect.cleandoc(
            """
            `acceleration` is passed to `optax.sgd` as the `nesterov` parameter.
            Note that this only has an effect if `momentum` is used as well.
            """
        )
        return inspect.cleandoc(note + "\n" + accel_nesterov)


class OptimistixOptaxLBFGS(AbstractOptimistixOptaxSolver):
    """
    L-BFGS implementation using optax.lbfgs wrapped by optimistix.OptaxMinimiser.

    Convergence criterion is implemented by Optimistix, so the Cauchy criterion is used.
    """

    fun: Callable
    fun_with_aux: Callable

    # stats: dict[str, PyTree[ArrayLike]]
    stats: dict[str, Pytree]

    _optax_solver = optax.lbfgs

    def __init__(
        self,
        unregularized_loss: Callable,
        regularizer: Regularizer,
        regularizer_strength: float | None,
        tol: float = DEFAULT_ATOL,
        rtol: float = DEFAULT_RTOL,
        maxiter: int = DEFAULT_MAX_STEPS,
        stepsize: float | None = None,
        memory_size: int = 10,
        scale_init_precond: bool = True,
        **solver_init_kwargs,
    ):
        """
        Create a solver wrapping `optax.lbfgs`.

        `stepsize` is passed for `learning_rate` to `optax.lbfgs`.
          If None, a zoom linesearch is used (recommended).
        `memory_size` and `scale_init_precond` are passed to `optax.lbfgs`.

        For more information on these parameters, see `get_solver_documentation`.
        """
        solver_init_kwargs["optim"] = optax.lbfgs(
            learning_rate=stepsize,
            memory_size=memory_size,
            scale_init_precond=scale_init_precond,
        )

        super().__init__(
            unregularized_loss,
            regularizer,
            regularizer_strength,
            tol=tol,
            rtol=rtol,
            maxiter=maxiter,
            **solver_init_kwargs,
        )


# fix optax.lbfgs doctest failing due to x64 being set in pytest
OptimistixOptaxLBFGS.__doc__ = (OptimistixOptaxLBFGS.__doc__ or "").replace(
    "jnp.array([1., 2., 3.])", "jnp.array([1., 2., 3.], dtype=jnp.float32)"
)



# required to get ArrayLike to render correctly
from __future__ import annotations

import abc
from typing import Optional, Tuple

import numpy as np
from numpy.typing import ArrayLike, NDArray
from pynapple import Tsd, TsdFrame, TsdTensor

from ..type_casting import support_pynapple
from ..typing import FeatureMatrix
from ._basis import Basis, check_transform_input, min_max_rescale_samples
from ._basis_mixin import AtomicBasisMixin


class RaisedCosineBasisLinear(AtomicBasisMixin, Basis, abc.ABC):
    r"""Represent linearly-spaced raised cosine basis functions.

    This implementation is based on the cosine bumps used by Pillow et al. [1]_
    to uniformly tile the internal points of the domain.

    Parameters
    ----------
    n_basis_funcs :
        The number of basis functions.
    width :
        Width of the raised cosine. Must be $x/2$ for any integer $x\geq 3$ (e.g., 1.5,
        2, 2.5, 3, etc.)
    label :
        The label of the basis, intended to be descriptive of the task variable being processed.
        For example: velocity, position, spike_counts.

    References
    ----------
    .. [1] Pillow, J. W., Paninski, L., Uzzel, V. J., Simoncelli, E. P., & J.,
        C. E. (2005). Prediction and decoding of retinal ganglion cell responses
        with a probabilistic spiking model. Journal of Neuroscience, 25(47),
        11003–11013.

    Raises
    ------
    ValueError
        If width <= 1 or 2*width is not a positive integer. Values that do not match
        this constraint will result in the following, which contribute to aliasing:
        - No overlap between bumps (width < 1).
        - Oscillatory behavior when summing the basis elements (2*width not integer).
    """

    def __init__(
        self,
        n_basis_funcs: int,
        width: float = 2.0,
        label: Optional[str] = "RaisedCosineBasisLinear",
    ) -> None:
        AtomicBasisMixin.__init__(self, n_basis_funcs=n_basis_funcs, label=label)
        Basis.__init__(
            self,
        )
        self._n_input_dimensionality = 1
        self._check_width(width)
        self._width = width
        # for these linear raised-cosine basis functions,
        # the samples must be rescaled to 0 and 1.
        self._rescale_samples = True

    @property
    def width(self):
        """Return width of the raised cosine."""
        return self._width

    @width.setter
    def width(self, width: float):
        self._check_width(width)
        self._width = width

    @staticmethod
    def _check_width(width: float) -> None:
        """Validate the width value.

        Parameters
        ----------
        width :
            The width value to validate.

        Raises
        ------
        ValueError
            If width <= 1 or 2*width is not a positive integer. Values that do not match
            this constraint will result in the following, which contribute to aliasing:
            - No overlap between bumps (width < 1).
            - Oscillatory behavior when summing the basis elements (2*width not integer).
        """
        if width <= 1 or (not np.isclose(width * 2, round(2 * width))):
            raise ValueError(
                f"Invalid raised cosine width. Width must be strictly greater than 1 and "
                f"2*width must be a integer, but got width={width} instead! At other "
                "values, basis sampling will lead to aliasing."
            )

    @support_pynapple(conv_type="numpy")
    @check_transform_input
    def evaluate(  # call these evaluate
        self,
        sample_pts: ArrayLike | Tsd | TsdFrame | TsdTensor,
    ) -> FeatureMatrix:
        """Evaluate the raised cosine basis at the sample points.

        Parameters
        ----------
        sample_pts :
            Spacing for basis functions, holding elements on interval [0, 1].
            `sample_pts` is a n-dimensional (n >= 1) array with first axis being the samples, i.e.
            `sample_pts.shape[0] == n_samples`.

        Raises
        ------
        ValueError
            If the sample provided do not lie in [0,1].

        """
        if self._rescale_samples:
            # note that sample points is converted to NDArray
            # with the decorator.
            # copy is necessary otherwise:
            # basis1 = nmo.basis.RaisedCosineBasisLinear(5)
            # basis2 = nmo.basis.RaisedCosineBasisLog(5)
            # additive_basis = basis1 + basis2
            # additive_basis.evaluate(*([x] * 2)) would modify both inputs
            sample_pts, _ = min_max_rescale_samples(
                np.copy(sample_pts), getattr(self, "bounds", None)
            )

        peaks = self._compute_peaks()
        delta = peaks[1] - peaks[0]

        # reshape samples
        shape = sample_pts.shape
        sample_pts = sample_pts.reshape(
            -1,
        )

        # generate a set of shifted cosines, and constrain them to be non-zero
        # over a single period, then enforce the codomain to be [0,1], by adding 1
        # and then multiply by 0.5
        basis_funcs = 0.5 * (
            np.cos(
                np.clip(
                    np.pi * (sample_pts[:, None] - peaks[None]) / (delta * self.width),
                    -np.pi,
                    np.pi,
                )
            )
            + 1
        )
        basis_funcs = basis_funcs.reshape(*shape, basis_funcs.shape[1])
        return basis_funcs

    def _compute_peaks(self) -> NDArray:
        """
        Compute the location of raised cosine peaks.

        Returns
        -------
            Peak locations of each basis element.
        """
        return np.linspace(0, 1, self.n_basis_funcs)

    def evaluate_on_grid(self, n_samples: int) -> Tuple[NDArray, NDArray]:
        """Evaluate the basis set on a grid of equi-spaced sample points.

        Parameters
        ----------
        n_samples :
            The number of points in the uniformly spaced grid. A higher number of
            samples will result in a more detailed visualization of the basis functions.

        Returns
        -------
        X :
            Array of shape (n_samples,) containing the equi-spaced sample
            points where we've evaluated the basis.
        basis_funcs :
            Raised cosine basis functions, shape (n_samples, n_basis_funcs)
        """
        return super().evaluate_on_grid(n_samples)

    def _check_n_basis_min(self) -> None:
        """Check that the user required enough basis elements.

        Check that the number of basis is at least 2.

        Raises
        ------
        ValueError
            If n_basis_funcs < 2.
        """
        if self.n_basis_funcs < 2:
            raise ValueError(
                f"Object class {self.__class__.__name__} requires >= 2 basis elements. "
                f"{self.n_basis_funcs} basis elements specified instead"
            )


class RaisedCosineBasisLog(RaisedCosineBasisLinear, abc.ABC):
    r"""Represent log-spaced raised cosine basis functions.

    Similar to ``RaisedCosineBasisLinear`` but the basis functions are log-spaced.
    This implementation is based on the cosine bumps used by Pillow et al. [1]_
    to uniformly tile the internal points of the domain.

    Parameters
    ----------
    n_basis_funcs :
        The number of basis functions.
    width :
        Width of the raised cosine. Must be $x/2$ for any integer $x\geq 3$ (e.g., 1.5,
        2, 2.5, 3, etc.)
    time_scaling :
        Non-negative hyper-parameter controlling the logarithmic stretch magnitude, with
        larger values resulting in more stretching. As this approaches 0, the
        transformation becomes linear.
    enforce_decay_to_zero:
        If set to True, the algorithm first constructs a basis with ``n_basis_funcs + ceil(width)`` elements
        and subsequently trims off the extra basis elements. This ensures that the final basis element
        decays to 0.
    label :
        The label of the basis, intended to be descriptive of the task variable being processed.
        For example: velocity, position, spike_counts.

    References
    ----------
    .. [1] Pillow, J. W., Paninski, L., Uzzel, V. J., Simoncelli, E. P., & J.,
       C. E. (2005). Prediction and decoding of retinal ganglion cell responses
       with a probabilistic spiking model. Journal of Neuroscience, 25(47),
       11003–11013.

    """

    def __init__(
        self,
        n_basis_funcs: int,
        width: float = 2.0,
        time_scaling: float = None,
        enforce_decay_to_zero: bool = True,
        label: Optional[str] = "RaisedCosineBasisLog",
    ) -> None:
        super().__init__(
            n_basis_funcs,
            width=width,
            label=label,
        )
        # The samples are scaled appropriately in the self._transform_samples which scales
        # and applies the log-stretch, no additional transform is needed.
        self._rescale_samples = False
        if time_scaling is None:
            time_scaling = 50.0

        self.time_scaling = time_scaling
        self.enforce_decay_to_zero = enforce_decay_to_zero

    @property
    def time_scaling(self):
        """Time scaling.

        Time scaling parameter regulating the logarithmic stretch of the basis.
        """
        return self._time_scaling

    @time_scaling.setter
    def time_scaling(self, time_scaling):
        """Setter property for time_scaling."""
        self._check_time_scaling(time_scaling)
        self._time_scaling = time_scaling

    @staticmethod
    def _check_time_scaling(time_scaling: float) -> None:
        if time_scaling <= 0:
            raise ValueError(
                f"Only strictly positive time_scaling are allowed, {time_scaling} provided instead."
            )

    def _transform_samples(
        self,
        sample_pts: ArrayLike,
    ) -> NDArray:
        """
        Map the sample domain to log-space.

        Parameters
        ----------
        sample_pts :
            Sample points used for evaluating the splines,
            shape (n_samples, ).

        Returns
        -------
            Transformed version of the sample points that matches the Raised Cosine basis domain,
            shape (n_samples, ).
        """
        # rescale to [0,1]
        # copy is necessary to avoid unwanted rescaling in additive/multiplicative basis.
        sample_pts, _ = min_max_rescale_samples(
            np.copy(sample_pts), getattr(self, "bounds", None)
        )
        # This log-stretching of the sample axis has the following effect:
        # - as the time_scaling tends to 0, the points will be linearly spaced across the whole domain.
        # - as the time_scaling tends to inf, basis will be small and dense around 0 and
        # progressively larger and less dense towards 1.
        log_spaced_pts = np.log(self.time_scaling * sample_pts + 1) / np.log(
            self.time_scaling + 1
        )
        return log_spaced_pts

    def _compute_peaks(self) -> NDArray:
        """
        Peak location of each log-spaced cosine basis element.

        Compute the peak location for the log-spaced raised cosine basis.
        Enforcing that the last basis decays to zero is equivalent to
        setting the last peak to a value smaller than 1.

        Returns
        -------
            Peak locations of each basis element.

        """
        if self.enforce_decay_to_zero:
            # compute the last peak location such that the last
            # basis element decays to zero at the last sample.
            last_peak = 1 - self.width / (self.n_basis_funcs + self.width - 1)
        else:
            last_peak = 1
        return np.linspace(0, last_peak, self.n_basis_funcs)

    @support_pynapple(conv_type="numpy")
    @check_transform_input
    def evaluate(
        self,
        sample_pts: ArrayLike | Tsd | TsdFrame | TsdTensor,
    ) -> FeatureMatrix:
        """Evaluate the log-spaced raised cosine basis at the sample points.

        Parameters
        ----------
        sample_pts :
            Spacing for basis functions. Samples will be rescaled to the interval [0, 1].
            `sample_pts` is a n-dimensional (n >= 1) array with first axis being the samples, i.e.
            `sample_pts.shape[0] == n_samples`.

        Returns
        -------
        basis_funcs :
            Log-raised cosine basis functions, shape (n_samples, n_basis_funcs).

        Raises
        ------
        ValueError
            If the sample provided do not lie in [0,1].
        """
        return super().evaluate(self._transform_samples(sample_pts))



"""
Module for fetching data and utilities using the pooch library.

This module allows you to download datasets and utility scripts
from specified URLs and manage their versions and integrity using
SHA256 hash verification. It also provides helper functions to
calculate hashes for local files and manage file paths.
"""

import os
import pathlib
from typing import Optional, Union

try:
    import pooch
    from pooch import Pooch
except ImportError:
    pooch = None
    Pooch = None

try:
    from tqdm.auto import tqdm
except ImportError:
    tqdm = None

try:
    import dandi
    import fsspec
    import h5py
    from dandi.dandiapi import DandiAPIClient
    from pynwb import NWBHDF5IO
except ImportError:
    dandi = None
    NWBHDF5IO = None

import hashlib

# Registry of dataset filenames and their corresponding SHA256 hashes.
REGISTRY_DATA = {
    "A0670-221213.nwb": "8587dd6dde107504bd4a17a68ce8fb934fcbcccc337e653f31484611ee51f50a",
    "Mouse32-140822.nwb": "1a919a033305b8f58b5c3e217577256183b14ed5b436d9c70989dee6dafe0f35",
    "Achilles_10252013.nwb": "42857015aad4c2f7f6f3d4022611a69bc86d714cf465183ce30955731e614990",
    "allen_478498617.nwb": "262393d7485a5b39cc80fb55011dcf21f86133f13d088e35439c2559fd4b49fa",
    "m691l1.nwb": "1990d8d95a70a29af95dade51e60ffae7a176f6207e80dbf9ccefaf418fe22b6",
    "A2929-200711.nwb": "f698d7319efa5dfeb18fb5fe718ec1a84fdf96b85a158177849a759cd5e396fe",
    "Achilles_10252013_EEG.nwb": "a97a69d231e7e91c07e24890225f8fe4636bac054de50345551f32fc46b9efdd",
    "em_three_states.npz": "92e9fe7990e98f3d23536a40658e258acbc83c26d773f78336431abc12e01951",
    "julia_regression_mstep_no_prior.npz": "412430a4e0d0bbc0abb224a3a8516f660266e3e5bc0d840c4ee6fd80b5501faa",
    "julia_regression_mstep_good_prior.npz": "e5bf87ca2f48b4904b0ce1c0e258f91c44033b05163e039f79b9b48c2c3e484d",
    "julia_regression_mstep_flat_prior.npz": "fdd1117c521945f25595f78f220256d37cb7bf37834403bd1cc04941e3203b22",
}
DOWNLOADABLE_FILES = list(REGISTRY_DATA.keys())

# URL templates for downloading datasets and utility scripts.
OSF_TEMPLATE = "https://osf.io/download/{}/"

# Mapping of dataset filenames to their download URLs.
REGISTRY_URLS_DATA = {
    "A0670-221213.nwb": OSF_TEMPLATE.format("sbnaw"),
    "Mouse32-140822.nwb": OSF_TEMPLATE.format("jb2gd"),
    "Achilles_10252013.nwb": OSF_TEMPLATE.format("hu5ma"),
    "allen_478498617.nwb": OSF_TEMPLATE.format("vf2nj"),
    "m691l1.nwb": OSF_TEMPLATE.format("xesdm"),
    "A2929-200711.nwb": OSF_TEMPLATE.format("y7zwd"),
    "Achilles_10252013_EEG.nwb": OSF_TEMPLATE.format("2dfvp"),
    "em_three_states.npz": OSF_TEMPLATE.format("wdz7j"),
    "julia_regression_mstep_no_prior.npz": OSF_TEMPLATE.format("3ky6m"),
    "julia_regression_mstep_good_prior.npz": OSF_TEMPLATE.format("w586p"),
    "julia_regression_mstep_flat_prior.npz": OSF_TEMPLATE.format("umz4d"),
}

_NEMOS_ENV = "NEMOS_DATA_DIR"


def _create_retriever(path: Optional[pathlib.Path] = None) -> Pooch:
    """Create a pooch retriever for fetching datasets.

    This function sets up the pooch retriever, which manages the
    downloading and caching of files, including handling retries
    and checking file integrity using SHA256 hashes.

    Parameters
    ----------
    path :
        The directory where datasets will be stored. If not provided,
        defaults to pooch's cache (check ``pooch.os_cache('nemos')`` for that path)

    Returns
    -------
    :
        A configured pooch retriever object.

    """
    if path is None:
        # Use the default data directory if none is provided.
        path = pooch.os_cache("nemos")

    return pooch.create(
        path=path,
        base_url="",
        urls=REGISTRY_URLS_DATA,
        registry=REGISTRY_DATA,
        retry_if_failed=2,
        allow_updates="POOCH_ALLOW_UPDATES",
        env=_NEMOS_ENV,
    )


def _check_dependencies(needs_dandi: bool = False):
    """Check optional dependencies."""
    if needs_dandi and dandi is None:
        raise ImportError(
            "Missing optional dependency 'dandi'."
            " Please use pip or "
            "conda to install 'dandi'."
        )

    if pooch is None and tqdm is None:
        raise ImportError(
            "Missing optional dependencies 'pooch' and 'tqdm'."
            " Please use pip or "
            "conda to install 'pooch' and 'tqdm'."
        )
    elif pooch is None:
        raise ImportError(
            "Missing optional dependency 'pooch'."
            " Please use pip or "
            "conda to install 'pooch'."
        )
    elif tqdm is None:
        raise ImportError(
            "Missing optional dependency 'tqdm'."
            " Please use pip or "
            "conda to install 'tqdm'."
        )


def fetch_data(
    dataset_name: str, path: Optional[Union[pathlib.Path, str]] = None
) -> str:
    """
    Download a dataset using pooch.

    This function downloads a dataset, checking if it already exists
    and is unchanged. If the dataset is an archive (ends in .tar.gz),
    it decompresses the archive and returns the path to the resulting
    directory. Otherwise, it returns the path to the downloaded file.

    Parameters
    ----------
    dataset_name :
        The name of the dataset to download. Must match an entry in
        REGISTRY_DATA.
    path :
        The directory where the dataset will be stored. If not provided,
        defaults to _DEFAULT_DATA_DIR.

    Returns
    -------
    :
        The path to the downloaded file or directory.
    """
    _check_dependencies()
    retriever = _create_retriever(path)
    # Fetch the dataset using pooch.
    return retriever.fetch(
        dataset_name,
        progressbar=True if tqdm else False,
    )


def download_dandi_data(
    dandiset_id: str, file_path: str, force_download: bool = False
) -> NWBHDF5IO:
    """Download a dataset from the [DANDI Archive](https://dandiarchive.org/).

    Parameters
    ----------
    dandiset_id :
        6-character string of numbers giving the ID of the dandiset.
    file_path :
        filepath to the specific .nwb file within the dandiset we wish to return.
    force_download :
        True if you want to download the dataset even if it already exists, False - default - otherwise.

    Returns
    -------
    io :
        NWB file containing specified data.

    Examples
    --------
    >>> import nemos as nmo
    >>> import pynapple as nap
    >>> io = nmo.fetch.download_dandi_data("000582",
    ...                                    "sub-11265/sub-11265_ses-07020602_behavior+ecephys.nwb")
    >>> nwb = nap.NWBFile(io.read(), lazy_loading=False)
    >>> print(nwb)
    07020602
    ┍━━━━━━━━━━━━━━━━━━━━━┯━━━━━━━━━━┑
    │ Keys                │ Type     │
    ┝━━━━━━━━━━━━━━━━━━━━━┿━━━━━━━━━━┥
    │ units               │ TsGroup  │
    │ ElectricalSeriesLFP │ Tsd      │
    │ SpatialSeriesLED2   │ TsdFrame │
    │ SpatialSeriesLED1   │ TsdFrame │
    │ ElectricalSeries    │ Tsd      │
    ┕━━━━━━━━━━━━━━━━━━━━━┷━━━━━━━━━━┙

    """
    _check_dependencies(needs_dandi=True)

    # Set up cache directory
    if _NEMOS_ENV in os.environ:
        cache_dir = pathlib.Path(os.environ[_NEMOS_ENV])
    else:
        cache_dir = pooch.os_cache("nemos") / "nwb-cache"
    cache_dir.mkdir(parents=True, exist_ok=True)

    # Create a deterministic filename based on dandiset_id and file_path
    # Hash to make sure that there are no problematic characters for filename.
    dandiset_hash = hashlib.md5(str(dandiset_id).encode()).hexdigest()
    filepath_hash = hashlib.md5(str(file_path).encode()).hexdigest()
    cache_filename = dandiset_hash + filepath_hash + ".nwb"
    cached_file_path = cache_dir / cache_filename

    # Check if file already exists in cache
    if cached_file_path.exists() and not force_download:
        # File exists, open it directly
        file = h5py.File(cached_file_path, "r")
        io = NWBHDF5IO(file=file, load_namespaces=True)
        return io

    # File doesn't exist, download it
    with DandiAPIClient() as client:
        asset = client.get_dandiset(dandiset_id, "draft").get_asset_by_path(file_path)
        s3_url = asset.get_content_url(follow_redirects=1, strip_query=True)

    # Download file using fsspec
    fs = fsspec.filesystem("http")

    # Download to temporary location first, then move to final location
    temp_file_path = cached_file_path.with_suffix(".tmp")

    try:
        # Get file size for progress bar
        with fs.open(s3_url, "rb") as remote_file:
            # Try to get content length from headers
            file_size = getattr(remote_file, "size", None)

        with fs.open(s3_url, "rb") as remote_file:
            # Initialize progress bar

            with tqdm(
                total=file_size,
                unit="B",
                unit_scale=True,
                unit_divisor=1024,
                desc=f"Downloading {file_path.split('/')[-1]}",
            ) as pbar:
                with open(temp_file_path, "wb") as local_file:
                    chunk_size = 8192 * 1024  # 8MB chunks
                    while True:
                        chunk = remote_file.read(chunk_size)
                        if not chunk:
                            break
                        local_file.write(chunk)
                        pbar.update(len(chunk))

        # Move completed download to final location
        temp_file_path.rename(cached_file_path)

    except Exception:
        # Clean up temp file if download failed
        if temp_file_path.exists():
            temp_file_path.unlink()
        raise

    # Open the downloaded file
    file = h5py.File(cached_file_path, "r")
    io = NWBHDF5IO(file=file, load_namespaces=True)

    return io



"""Module for ND fourier basis class."""

import itertools
import warnings
from numbers import Number
from typing import Any, Callable, List, Literal, Optional, Tuple

import jax
import jax.numpy as jnp
import numpy as np
from numpy._typing import NDArray
from numpy.typing import ArrayLike
from pynapple import Tsd, TsdFrame, TsdTensor

from ..type_casting import is_at_least_1d_numpy_array_like, support_pynapple
from ..typing import FeatureMatrix
from ._basis import Basis, check_transform_input, min_max_rescale_samples
from ._basis_mixin import AtomicBasisMixin

FREQUENCY_ERROR_MSGS = {
    "has_negative_values": "The provided frequencies contain negative values. Valid frequencies must be"
    "non-negative integers.",
    "not_int_like": "Some frequency values are not integers. Valid frequencies must be"
    "non-negative integers.",
}


def _check_array_properties(*arrays: ArrayLike):
    is_int_like = all(jnp.all(f == jnp.asarray(f, dtype=int)) for f in arrays)
    is_positive = all(jnp.all(jnp.asarray(f) >= 0) for f in arrays)
    return is_int_like, is_positive


def _format_error_message(
    is_positive: bool,
    is_int_like: bool,
):
    msg = "Invalid frequencies provided. The following issues were detected:\n\n"

    if not is_positive:
        msg += (
            "\t- "
            + FREQUENCY_ERROR_MSGS["has_negative_values"].replace("\n", "\n\t  ")
            + "\n\n"
        )
    if not is_int_like:
        msg += (
            "\t- "
            + FREQUENCY_ERROR_MSGS["not_int_like"].replace("\n", "\n\t  ")
            + "\n\n"
        )

    return msg


def _check_and_sort_frequencies(*frequencies) -> List[jnp.ndarray]:
    is_int_like, is_positive = _check_array_properties(*frequencies)

    if is_int_like and is_positive:
        sorted_freqs = [jnp.sort(f).astype(float) for f in frequencies]
        if any(
            not jnp.array_equal(f1, f2) for f1, f2 in zip(frequencies, sorted_freqs)
        ):
            warnings.warn("Unsorted frequencies provided! Frequencies will be sorted.")
        return sorted_freqs

    raise ValueError(_format_error_message(is_positive, is_int_like))


def arange_constructor(arg: NDArray | int | Tuple[int, int]):
    """
    Create an array of frequencies from different types of input.

    This function accepts either:
    - An array-like object of frequencies (returned after sorting and checking),
    - A single positive integer (returns an array from 0 to `arg - 1`),
    - A tuple of two integers `(start, stop)` (returns an array from `start` to `stop - 1`).

    Parameters
    ----------
    arg :
        Specifies how to create the frequency array:
        * If array-like, it is validated and sorted before returning.
        * If int, must be > 0, and the result is `jnp.arange(arg, dtype=float)`.
        * If tuple of two ints `(start, stop)`, must satisfy `0 <= start < stop`,
          and the result is `jnp.arange(start, stop, dtype=float)`.

    Returns
    -------
    :
        A 1D array of frequencies as floats.

    Raises
    ------
    ValueError
        If the integer input is not positive, or if the tuple does not satisfy
        `0 <= start < stop`.
    TypeError
        If the tuple elements are not integers or if the input is of an unsupported type.

    Notes
    -----
    If an array-like input is provided, it is first validated and sorted using
    `_check_and_sort_frequencies`.

    Examples
    --------
    >>> arange_constructor(3)
    Array([0., 1., 2.], dtype=float32)

    >>> arange_constructor((2, 5))
    Array([2., 3., 4.], dtype=float32)

    >>> arange_constructor(jnp.array([3, 1, 2]))
    Array([1., 2., 3.], dtype=float32)
    """
    if is_at_least_1d_numpy_array_like(arg):
        arg = _check_and_sort_frequencies(arg)
        return arg[0]

    if isinstance(arg, int):
        if arg < 0:
            raise ValueError(
                f"Integer frequencies must be >= 0, {arg} provided instead."
            )
        return jnp.arange(arg, dtype=float)

    elif isinstance(arg, tuple) and len(arg) == 2:
        start, stop = arg
        if not (isinstance(start, int) and isinstance(stop, int)):
            raise TypeError("Tuple frequencies must be integers.")
        if start < 0 or stop <= start:
            raise ValueError(
                f"Tuple frequencies must satisfy 0 <= start < stop. "
                f"Start is ``{start}`` and stop is ``{stop}`` instead."
            )
        return jnp.arange(start, stop, dtype=float)
    else:
        raise TypeError("Each frequency must be an int or a 2-element tuple of ints.")


def _process_tuple_frequencies(frequencies: tuple, ndim: int):
    """
    Process a tuple of frequencies and return the corresponding frequency arrays.

    This function handles two valid cases for specifying frequencies as a tuple:

    - A 2-element tuple of integers:
       - Both elements must be numeric and equal to their integer representation.
       - Returns a single-element tuple containing the result of `arange_constructor(frequencies)`.


    Parameters
    ----------
    frequencies :
        The input tuple specifying frequencies. It must be a 2-element tuple of integers.

    ndim :
        The expected number of input dimensions. Used to validate the tuple length
        in the second case.

    Returns
    -------
    :
        A tuple of frequency arrays processed according to the input.

    Raises
    ------
    ValueError
        If the tuple does not match one of the expected formats.

    Examples
    --------
    >>> _process_tuple_frequencies((2, 5), 1)
    [Array([2., 3., 4.], dtype=float32)]

    """
    if len(frequencies) == 2 and all(
        isinstance(f, Number) and (f == int(f)) for f in frequencies
    ):
        return [arange_constructor(frequencies)]

    raise ValueError(
        "Invalid frequencies specification. If ``frequencies`` are provided as a tuple, "
        f"it must be a 2-element tuple of non-negative integers. Tuple {frequencies} provided instead."
    )


def _get_all_frequency_pairs(frequencies):
    grids = jnp.meshgrid(*[freqs for freqs in frequencies], indexing="ij")
    return jnp.stack([g.reshape(-1) for g in grids])


def _get_frequency_pairs_from_callable(
    frequency_mask: Callable[..., bool], frequencies: Tuple[jnp.ndarray, ...]
):
    """
    Apply the callable assigned to `frequency_mask` to all frequency tuples.

    Parameters
    ----------
    frequency_mask :
        A function with signature: frequency_mask(*freqs) -> bool (or 0/1).
    frequencies :
        1D arrays of frequencies (one per dimension).

    Returns
    -------
    :
        Shape (D, K): columns are the selected frequency tuples.
    """
    all_pairs = itertools.product(*frequencies)  # shape (D, N)

    if not callable(frequency_mask):
        raise TypeError(
            "`frequency_mask` must be a callable like frequency_mask(*freqs) -> bool."
        )

    selected = []
    for j, freqs in enumerate(all_pairs):
        try:
            include = frequency_mask(*freqs)
        except Exception as e:
            raise TypeError(
                "Error while applying the callable assigned to `frequency_mask`.\n"
                "Expected signature: frequency_mask(*frequencies) -> bool.\n"
                f"Failed at index {j} with frequencies={freqs!r}."
            ) from e

        # Normalize/validate the result to a single boolean
        is_valid_number = isinstance(include, (Number, np.bool_)) and include in (0, 1)
        is_valid_array = (
            isinstance(include, (jnp.ndarray, np.ndarray))
            and include.size == 1
            and include in (0, 1)
        )
        if is_valid_number or is_valid_array:
            include = bool(include)
        else:
            string_vals = ", ".join(str(f) for f in freqs)
            raise ValueError(
                "`frequency_mask(*freqs)` must return a single boolean or 0/1.\n"
                f"``frequency_mask({string_vals})`` returned {include!r} "
                f"of type {type(include).__name__}."
            )

        if include:
            selected.append(jnp.asarray(freqs))

    return (
        jnp.stack(selected, axis=1)
        if selected
        else jnp.zeros((len(frequencies), 0), dtype=int)
    )


class FourierBasis(AtomicBasisMixin, Basis):
    """
    N-dimensional Fourier basis for feature expansion.

    This class generates a set of sine and cosine basis functions defined over
    an ``n``-dimensional input space. The basis functions are constructed from
    a Cartesian product of frequencies specified for each input dimension.
    Each selected frequency combination contributes two basis functions
    (cosine and sine), except for the all-zero frequency (DC component),
    which contributes only a cosine term.

    The class supports flexible frequency specification (integers, ranges, or
    lists per dimension) and optional masking to include or exclude specific
    frequency combinations.

    Parameters
    ----------
    frequencies :
        Frequency specification(s).

        **Single specification** (broadcasted to all dimensions when ``ndim > 1``):
          * ``int k`` with ``k >= 0``.
          * ``(low, high)``: a 2-element tuple of integers with ``0 <= low < high``.
          * 1-D NumPy ``ndarray`` of non-negative integers. If not sorted ascending,
            a ``UserWarning`` is issued.

        **Per-dimension container**:
          * A **list** of length ``ndim`` whose elements are each a valid single specification.
            For ``ndim == 1``, a length-1 list is also accepted.

    ndim :
        Dimensionality of the basis. Default is 1.

    frequency_mask :
        Frequency selection mask. Used to filter the evaluated frequencies after
        construction.

        Accepted forms:
            * ``no-intercept``: drop the all-zero frequency (DC component).
            * ``"all"``: Keep all frequencies
            * **None** : Keep all frequencies.
            * **array_like** of {0, 1} or booleans : Shape must match the number
              of possible frequency combinations for the given ``frequencies``.
              - In 1D: shape = (n_freq,).
              - In nD: shape = (n_freq_dim1, n_freq_dim2, ..., n_freq_dimN).
            * **callable** : A function applied to each tuple of frequency
              coordinates, returning a single boolean or {0, 1}. For example:
              ``lambda f1, f2, ...: condition``.
                - Must return a scalar boolean or integer {0, 1}.
                - Returning arrays, lists, or non-boolean values raises a
                  ``ValueError``.

        Validation rules:
            * Array values must be exactly 0 or 1 (floats are allowed if equal to
              0.0 or 1.0).
            * Strings or non-numeric values raise ``ValueError``.
            * Callable return values must be a single boolean or {0, 1}; anything
              else raises ``ValueError``.
            * Errors raised inside the callable are propagated as
              ``TypeError`` with a descriptive message.

    label : str, optional
        Descriptive label for the basis (e.g., to use in plots or summaries).

    Notes
    -----
    - If ``frequency_mask`` is provided, only the selected frequency
      combinations are used to build the basis.
    - The output of ``compute_features`` contains both cosine and sine components for
      each active frequency combination, except that the all-zero frequency
      includes only a cosine term.
    - When a **tuple** is provided as a frequency, it is interpreted
      as a single range specification. Tuples that are not exactly a 2-element
      tuple of non-negative integers are invalid.
    """

    _is_complex = True

    def __init__(
        self,
        frequencies: (
            int
            | Tuple[int, int]
            | List[int]
            | List[Tuple[int, int]]
            | NDArray
            | List[NDArray]
        ),
        ndim: int,
        frequency_mask: (
            Literal["all", "no-intercept"] | jnp.ndarray | None
        ) = "no-intercept",
        label: Optional[str] = None,
    ) -> None:
        self._n_input_dimensionality = self._check_ndim(ndim)
        self.frequencies = frequencies
        self.frequency_mask = frequency_mask

        AtomicBasisMixin.__init__(
            self,
            n_basis_funcs=self.n_basis_funcs,
            label=label,
        )

    @property
    def frequency_mask(self) -> Callable | jnp.ndarray | Literal["all", "no-intercept"]:
        """Get the frequency mask for the Fourier basis.

        The frequency mask can be either:

        - a boolean array (or array-like of 0s and 1s) whose shape matches the number
          of frequencies along each input dimension, or
        - a callable with signature ``frequency_mask(*freqs) -> bool`` (or 0/1) applied
          to each frequency tuple, or
        - a string, either ``"all"``, if all possible frequency combinations are included, or
          ``"no-intercept"``, if the intercept term is dropped (DC component).

        Returns
        -------
        :
            The callable used to build the mask, the boolean JAX array mask,
            or ``None`` if no mask is applied.
        """
        # safe get when getter is called at init initialization
        return getattr(self, "_frequency_mask", "no-intercept")

    @frequency_mask.setter
    def frequency_mask(
        self,
        values: (
            Literal["all", "no-intercept"]
            | ArrayLike
            | jnp.ndarray
            | Callable[..., bool]
            | None
        ),
    ) -> None:
        """Set the frequency mask for the Fourier basis.

        Parameters
        ----------
        values :
            One of:
            - :class:`Literal <typing.Literal>`: either `"no-intercept"`` - default - which drops
              the 0-frequency DC term, or ``"all"`` which keeps all the frequencies -
              equivalent to :class:`None <NoneType>`.
            - **Array / array-like (bool or 0/1)**: Explicit mask over the
              frequency grid. Must have shape
              ``(len(frequencies[0]), len(frequencies[1]), ...)`` and contain
              only booleans or the integers {0, 1}.
            - :class:`callable`: A function with signature
              ``frequency_mask(*freqs) -> bool`` (or 0/1). It is applied to each
              frequency tuple ``(f1, f2, ..., f_n)`` to build the mask. The callable is
              **not required to be vectorized**; ``n`` is the input dimensionality.
            - :class:`None <NoneType>`: Include all frequency combinations.

        Raises
        ------
        ValueError
            If an array mask has values other than {0, 1}/booleans, or if its
            shape does not match the expected grid shape.
        TypeError
            If ``values`` is neither array-like, callable, nor ``None``; or if
            the callable returns a value that is not a single boolean or 0/1.

        Notes
        -----
        - Setting this property updates:
          ``self._frequency_mask`` (callable or boolean mask),
          ``self._n_basis_funcs`` (number of active basis functions),
          and ``self._eval_freq`` (selected frequencies for evaluation).
        """

        if isinstance(values, str) and values == "no-intercept":
            self._frequency_mask = "no-intercept"
            self._freq_combinations = _get_all_frequency_pairs(self._frequencies)
            if jnp.all(self._freq_combinations[..., 0] == 0):
                self._freq_combinations = self._freq_combinations[..., 1:]

        elif values is None or isinstance(values, str) and values == "all":
            self._frequency_mask = "all"
            self._freq_combinations = _get_all_frequency_pairs(self._frequencies)

        elif callable(values):
            self._freq_combinations = _get_frequency_pairs_from_callable(
                values, self._frequencies
            )
            self._frequency_mask = values
        else:
            try:
                values = jnp.asarray(values)
            except Exception as e:
                raise ValueError(
                    f"``frequency_mask`` {values} cannot be converted to a jax array of boolean."
                ) from e

            if not jnp.all((values == 0) | (values == 1)):
                raise ValueError("Frequency mask must be an array-like of 0s and 1s.")

            values = values.astype(bool)

            if not values.ndim == self._n_input_dimensionality:
                ndim = self._n_input_dimensionality
                raise ValueError(
                    f"The frequency mask for a {ndim}-dimensional Fourier basis "
                    f"must be an {ndim}-dimensional array of 0s and 1s. "
                    f"The provided mask is an {ndim}-dimensional array instead."
                )

            if tuple(len(freqs) for freqs in self._frequencies) != values.shape:
                expected_shape = tuple(len(freqs) for freqs in self._frequencies)
                raise ValueError(
                    "Invalid shape for ``frequency_mask``. "
                    f"Expected shape {expected_shape}, "
                    f"but got {values.shape} instead. The mask must have one entry (0 or 1) for each "
                    "frequency along every dimension of the Fourier basis."
                )

            self._frequency_mask = values
            self._n_basis_funcs = 2 * int(jnp.sum(self._frequency_mask)) - int(
                self._frequency_mask[(0,) * values.ndim]
            )

            idxs = jnp.stack(jnp.where(self._frequency_mask))
            self._freq_combinations = jnp.stack(
                [freqs[idxs[d]] for d, freqs in enumerate(self._frequencies)]
            )

        # used to drop or not the zero phase
        self._has_zero_phase = (
            0
            if self._freq_combinations.size == 0
            else int(jnp.all(self._freq_combinations[:, 0] == 0))
        )

    @property
    def frequencies(self) -> List[jnp.ndarray]:
        """Frequencies for the basis.

        Returns
        -------
        :
            A tuple of arrays with the fourier frequencies, one per
            dimension of the basis.
        """
        return self._frequencies

    @frequencies.setter
    def frequencies(
        self,
        frequencies: int | tuple[int, int] | list[int] | list[tuple[int, int]],
    ) -> None:
        ndim = self._n_input_dimensionality

        if isinstance(frequencies, Number) and (frequencies == int(frequencies)):
            frequencies = [arange_constructor(frequencies)] * ndim

        elif isinstance(frequencies, tuple):
            frequencies = _process_tuple_frequencies(
                frequencies, self._n_input_dimensionality
            )

        elif is_at_least_1d_numpy_array_like(frequencies):
            frequencies = _check_and_sort_frequencies(*([frequencies] * ndim))

        elif isinstance(frequencies, list):
            if len(frequencies) != self._n_input_dimensionality:
                raise ValueError(
                    "Length of frequencies list must match input dimensionality."
                )
            frequencies = [arange_constructor(f) for f in frequencies]

        else:
            if isinstance(frequencies, (np.ndarray, jnp.ndarray)) and ~np.issubdtype(
                frequencies.dtype, np.integer
            ):
                type_string = f"NDArray[{frequencies.dtype}]"
            else:
                type_string = repr(type(frequencies))
            raise TypeError(
                f"Unrecognized type {type_string} for the ``frequencies`` parameter. ``frequencies`` "
                "must be one of:\n\n"
                "  - int\n"
                "  - tuple[int, int]\n"
                "  - NDArray[int]\n"
                "  - list[int | tuple[int, int] | NDArray[int] | NDArray[int]]\n\n"
                f"If a list is provided, the list should be of length ``{ndim}``, "
                "one entry for each dimension of the Fourier basis."
            )

        # Skip update if same as current
        current_freqs = getattr(self, "_frequencies", [])
        if len(frequencies) == len(current_freqs) and all(
            np.array_equal(f1, f2) for f1, f2 in zip(current_freqs, frequencies)
        ):
            return

        self._frequencies = frequencies

        if not isinstance(self.frequency_mask, str):
            warnings.warn(
                "Resetting ``frequency_mask`` to ``'no-intercept'`` (all frequencies "
                "except the intercept - DC term - will be included).\n"
                "To sub-select frequencies, please provide a new ``frequency_mask``.",
                UserWarning,
            )
            self.frequency_mask = "no-intercept"
        else:
            # call the setter to re-calculate frequency pairs
            self.frequency_mask = self.frequency_mask

    @property
    def masked_frequencies(self) -> jnp.ndarray:
        """
        The frequencies after the masking is applied.

        Returns
        -------
            The masked frequencies, shape ``(ndim, n_frequency_combinations)``.
            ``masked_frequencies[:, i]`` is the frequency combination for the
            i-th basis function.

        """
        return self._freq_combinations

    @property
    def ndim(self):
        """The dimensionality of the basis."""
        return self._n_input_dimensionality

    @staticmethod
    def _check_ndim(ndim: int) -> int:
        try:
            is_int = int(ndim) == ndim
        except Exception as e:
            raise TypeError(f"Cannot convert ndim {ndim!r} to type int.") from e
        is_positive = ndim > 0
        if not is_int or not is_positive:
            raise ValueError(
                f"ndim must be a non-negative integer. {ndim!r} provided instead."
            )
        return int(ndim)

    @property
    def n_basis_funcs(self) -> int | None:
        return 2 * self._freq_combinations.shape[-1] - self._has_zero_phase

    @support_pynapple(conv_type="numpy")
    @check_transform_input
    def evaluate(  # call these _evaluate
        self,
        *sample_pts: ArrayLike | Tsd | TsdFrame | TsdTensor,
    ) -> FeatureMatrix:
        """Evaluate the Fourier basis at the sample points.

        Parameters
        ----------
        sample_pts :
            Spacing for basis functions, holding elements on interval [0, 1].
            `sample_pts` is a n-dimensional (n >= 1) array with first axis being the samples, i.e.
            `sample_pts.shape[0] == n_samples`.

        Raises
        ------
        ValueError
            If the sample provided do not lie in [0,1].

        """
        shape = sample_pts[0].shape

        bounds = getattr(self, "bounds", None)
        if bounds is None:
            bounds = (None,) * self._n_input_dimensionality

        # min/max rescale to [0,1]:
        # The function does so over the time axis (each extra dim is
        # normalized independently)
        def _flat_samples_to_angles(xs):
            scaled_samples = jax.tree_util.tree_map(
                lambda x, b: 2
                * jnp.pi
                * self._shift_angles(min_max_rescale_samples(x, b)[0].reshape(-1)),
                xs,
                bounds,
            )
            return jnp.stack(scaled_samples, axis=-1)

        sample_pts = _flat_samples_to_angles(sample_pts)
        angles = sample_pts @ self._freq_combinations
        out = jnp.concatenate(
            [jnp.cos(angles), jnp.sin(angles[..., self._has_zero_phase :])], axis=1
        )
        return out.reshape(*shape, out.shape[-1])

    def evaluate_on_grid(self, *n_samples: int) -> Tuple[Tuple[NDArray], NDArray]:
        """Evaluate the basis set on a grid of equi-spaced sample points.

        Parameters
        ----------
        n_samples :
            The number of points in the uniformly spaced grid. A higher number of
            samples will result in a more detailed visualization of the basis functions.

        Returns
        -------
        X :
            Array of shape (n_samples,) containing the equi-spaced sample
            points where we've evaluated the basis.
        basis_funcs :
            Fourier basis functions, shape (n_samples, n_basis_funcs)
        """
        return super().evaluate_on_grid(*n_samples)

    def _shift_angles(self, sample_pts: ArrayLike) -> ArrayLike:
        """
        Shift angles.

        Reimplemented for ``FourierConv``, shifting the angles to
        match the Fourier coefficients when the basis is used for convolutions.
        This shift must not be applied for ``FourierEval`` basis, therefore the
        super-class implements an identity function.

        Parameters
        ----------
        sample_pts :
            The samples.

        Returns
        -------
        sample_pts :
            The samples as provided, identity function.
        """
        return sample_pts

    def set_params(self, **params: Any):
        """Set params handling correctly the frequencies and their mask."""
        # if both frequencies and mask are set ignore warning
        if "frequencies" in params and "frequency_mask" in params:
            freq = params.pop("frequencies")
            with warnings.catch_warnings():
                warnings.filterwarnings(
                    "ignore",
                    category=UserWarning,
                    message="Resetting ``frequency_mask``.*",
                )
                # set first frequencies
                self.frequencies = freq
                # then set everything else (so that the mask is
                # checked against the new frequencies)
                return super().set_params(**params)
        else:
            return super().set_params(**params)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from absl.testing import absltest
from absl.testing import parameterized

import jax
import jax.numpy as jnp

from nemos.third_party.jaxopt.jaxopt import GaussNewton
from nemos.third_party.jaxopt.jaxopt._src import test_util

import numpy as onp


def _enzyme_reaction_residual_model(coeffs, x, y):
  return y - coeffs[0] * x / (coeffs[1] + x)


def _enzyme_reaction_residual_model_jac(coeffs, x, y, eps=1e-5):
  """Return the numerical Jacobian."""
  gn = GaussNewton(
      residual_fun=_enzyme_reaction_residual_model,
      maxiter=100,
      tol=1.0e-6)

  # Sets eps only at idx, the rest is zero
  eps_at = lambda idx: onp.array([int(i == idx)*eps for i in range(len(x))])

  res1 = jnp.zeros((len(coeffs), len(x)))
  res2 = jnp.zeros((len(coeffs), len(x)))
  for i in range(len(x)):
    res1 = res1.at[:,i].set(gn.run(coeffs, x + eps_at(i), y).params)
    res2 = res2.at[:,i].set(gn.run(coeffs, x - eps_at(i), y).params)

  twoeps = 2 * eps
  return (res1 - res2) / twoeps


def _city_temperature_residual_model(coeffs, x, y):
  return y - (coeffs[0] * jnp.sin(x * coeffs[1] + coeffs[2]) + coeffs[3])


class GaussNewtonTest(test_util.JaxoptTestCase):

  def setUp(self):
    super().setUp()

    self.substrate_conc = onp.array(
        [0.038, 0.194, .425, .626, 1.253, 2.500, 3.740])
    self.rate_data = onp.array(
        [0.050, 0.127, 0.094, 0.2122, 0.2729, 0.2665, 0.3317])
    self.init_enzyme_reaction_coeffs = onp.array([0.1, 0.1])

    self.months = onp.arange(1, 13)
    self.temperature_record = onp.array([
        61.0, 65.0, 72.0, 78.0, 85.0, 90.0, 92.0, 92.0, 88.0, 81.0, 72.0, 63.0
    ])
    self.init_temperature_record_coeffs = onp.array([10, 0.5, 10.5, 50])

  def test_aux_true(self):
    gn = GaussNewton(lambda x: (x**2, True), has_aux=True, maxiter=2)
    x_init = jnp.arange(2.)
    _, state = gn.run(x_init)
    self.assertEqual(state.aux, True)

  # Example taken from "Probability, Statistics and Estimation" by Mathieu ROUAUD. 
  # The algorithm is detailed and applied to the biology experiment discussed in 
  # page 84 with the uncertainties on the estimated values.
  def test_enzyme_reaction_parameter_fit(self):
    gn = GaussNewton(
        residual_fun=_enzyme_reaction_residual_model,
        maxiter=100,
        tol=1.0e-6)
    optimize_info = gn.run(
        self.init_enzyme_reaction_coeffs,
        self.substrate_conc,
        self.rate_data)

    self.assertArraysAllClose(optimize_info.params,
                              onp.array([0.36183689, 0.55626653]), 
                              rtol=1e-7, atol=1e-7)

  @parameterized.product(implicit_diff=[True, False])
  def test_enzyme_reaction_implicit_diff(self, implicit_diff):
    jac_num = _enzyme_reaction_residual_model_jac(
        self.init_enzyme_reaction_coeffs, self.substrate_conc, self.rate_data)

    gn = GaussNewton(
        residual_fun=_enzyme_reaction_residual_model,
        tol=1.0e-6,
        maxiter=10,
        implicit_diff=implicit_diff)

    def wrapper(substrate_conc):
      return gn.run(
          self.init_enzyme_reaction_coeffs,
          substrate_conc,
          self.rate_data).params
    jac_custom = jax.jacrev(wrapper)(self.substrate_conc)

    self.assertArraysAllClose(jac_num, jac_custom, atol=1e-2)

  # Example 7 from "SOLVING NONLINEAR LEAST-SQUARES PROBLEMS WITH THE 
  # GAUSS-NEWTON AND LEVENBERG-MARQUARDT METHODS" by ALFONSO CROEZE et al. 
  def test_temperature_record_four_parameter_fit(self):
    gn = GaussNewton(
        residual_fun=_city_temperature_residual_model,
        tol=1.0e-6)
    optimize_info = gn.run(
        self.init_temperature_record_coeffs,
        self.months,
        self.temperature_record)

    # Checking against the expected values
    self.assertArraysAllClose(
        optimize_info.params,
        onp.array([16.63994555, 0.46327812, 10.85228919, 76.19086103]),
        rtol=1e-6, atol=1e-5)

  def test_scalar_output_fun(self):
    gn = GaussNewton(
        residual_fun=lambda x: x @ x,
        tol=1e-1,)
    x_init = jnp.ones((2,))
    x_opt, _ = gn.run(x_init)

    self.assertAllClose(x_opt, jnp.zeros((2,)), atol=1e0)


if __name__ == '__main__':
  absltest.main()
